#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import print_function

import os
import sys
import glob
import datetime
from optparse import OptionParser

from paleomix.common.console import \
    print_info, \
    print_err


_TEMPLATE_TOP = \
    """# -*- mode: Yaml; -*-
# Timestamp: %s
#
# Default options.
# Can also be specific for a set of samples, libraries, and lanes,
# by including the "Options" hierarchy at the same level as those
# samples, libraries, or lanes below. This does not include
# "Features", which may only be specific globally.
Options:
  # Sequencing platform, see SAM/BAM reference for valid values
  Platform: Illumina
  # Quality offset for Phred scores, either 33 (Sanger/Illumina 1.8+)
  # or 64 (Illumina 1.3+ / 1.5+). For Bowtie2 it is also possible to
  # specify 'Solexa', to handle reads on the Solexa scale. This is
  # used during adapter-trimming and sequence alignment
  QualityOffset: 33
  # Split a lane into multiple entries, one for each (pair of) file(s)
  # found using the search-string specified for a given lane. Each
  # lane is named by adding a number to the end of the given barcode.
  SplitLanesByFilenames: yes
  # Compression format for FASTQ reads; 'gz' for GZip, 'bz2' for BZip2
  CompressionFormat: bz2

  # Settings for trimming of reads, see AdapterRemoval man-page
  AdapterRemoval:
     # Adapter sequences, set and uncomment to override defaults
#     --adapter1: AGATCGGAAGAGCACACGTCTGAACTCCAGTCACNNNNNNATCTCGTATGCCGTCTTCTGCTTG
#     --adapter2: AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT
     # Some BAM pipeline defaults differ from AR defaults;
     # To override, change these value(s):
     --mm: 3
     --minlength: 25
     # Extra features enabled by default; change 'yes' to 'no' to disable
     --collapse: yes
     --trimns: yes
     --trimqualities: yes
"""

_TEMPLATE_BAM_OPTIONS = \
    """  # Settings for aligners supported by the pipeline
  Aligners:
    # Choice of aligner software to use, either "BWA" or "Bowtie2"
    Program: BWA

    # Settings for mappings performed using BWA
    BWA:
      # One of "backtrack", "bwasw", or "mem"; see the BWA documentation
      # for a description of each algorithm (defaults to 'backtrack')
      Algorithm: backtrack
      # Filter aligned reads with a mapping quality (Phred) below this value
      MinQuality: 0
      # Filter reads that did not map to the reference sequence
      FilterUnmappedReads: yes
      # May be disabled ("no") for aDNA alignments, as post-mortem damage
      # localizes to the seed region, which BWA expects to have few
      # errors (sets "-l"). See http://pmid.us/22574660
      UseSeed: yes
      # Additional command-line options may be specified for the "aln"
      # call(s), as described below for Bowtie2 below.

    # Settings for mappings performed using Bowtie2
    Bowtie2:
      # Filter aligned reads with a mapping quality (Phred) below this value
      MinQuality: 0
      # Filter reads that did not map to the reference sequence
      FilterUnmappedReads: yes
      # Examples of how to add additional command-line options
#      --trim5: 5
#      --trim3: 5
      # Note that the colon is required, even if no value is specified
      --very-sensitive:
      # Example of how to specify multiple values for an option
#      --rg:
#        - CN:SequencingCenterNameHere
#        - DS:DescriptionOfReadGroup

  # Mark / filter PCR duplicates. If set to 'filter', PCR duplicates are
  # removed from the output files; if set to 'mark', PCR duplicates are
  # flagged with bit 0x400, and not removed from the output files; if set to
  # 'no', the reads are assumed to not have been amplified. Collapsed reads
  # are filtered using the command 'paleomix rmdup_duplicates', while "normal"
  # reads are filtered using Picard MarkDuplicates.
  PCRDuplicates: filter

  # Carry out quality base re-scaling of libraries using mapDamage
  # This will be done using the options set for mapDamage below
  RescaleQualities: no

  # Command-line options for mapDamage; note that the long-form
  # options are expected; --length, not -l, etc. Uncomment the
  # "mapDamage" line adding command-line options below.
  mapDamage:
    # By default, the pipeline will downsample the input to 100k hits
    # when running mapDamage; remove to use all hits
    --downsample: 100000

  # Set to 'yes' exclude a type of trimmed reads from alignment / analysis;
  # possible read-types reflect the output of AdapterRemoval
  ExcludeReads:
    Single: no              # Single-ended reads / Orphaned paired-ended reads
    Paired: no              # Paired ended reads
    Singleton: no           # Paired reads for which the mate was discarded
    Collapsed: no           # Overlapping paired-ended reads collapsed into a
                            # single sequence by AdapterRemoval
    CollapsedTruncated: no  # Like 'Collapsed', except that the reads
                            # truncated due to the presence ambigious
                            # bases or low quality bases at read termini.

  # Optional steps to perform during processing
  Features:
    RawBAM: no          # Generate BAM from the raw libraries (no indel realignment)
                        #   Location: {Destination}/{Target}.{Genome}.bam
    RealignedBAM: yes   # Generate indel-realigned BAM using the GATK Indel realigner
                        #   Location: {Destination}/{Target}.{Genome}.realigned.bam
    mapDamage: yes      # Generate mapDamage plot for each (unrealigned) library
                        #   Location: {Destination}/{Target}.{Genome}.mapDamage/{Library}/
    Coverage: yes       # Generate coverage information for the raw BAM (wo/ indel realignment)
                        #   Location: {Destination}/{Target}.{Genome}.coverage
    Depths: yes         # Generate histogram of number of sites with a given read-depth
                        #   Location: {Destination}/{Target}.{Genome}.depths
    Summary: yes        # Generate summary table for each target
                        #   Location: {Destination}/{Target}.summary
    DuplicateHist: no   # Generate histogram of PCR duplicates, for use with PreSeq
                        #   Location: {Destination}/{Target}.{Genome}.duphist/{Library}/
"""

_TEMPLATE_PREFIXES = """
# Map of prefixes by name, each having a Path key, which specifies the
# location of the BWA/Bowtie2 index, and optional label, and an option
# set of regions for which additional statistics are produced.
Prefixes:
  # Name of the prefix; is used as part of the output filenames
  NAME_OF_PREFIX:
    # Path to .fasta file containg a set of reference sequences.
    Path: PATH_TO_PREFIX

    # Label for prefix: One of nuclear, mitochondrial, chloroplast,
    # plasmid, bacterial, or viral. Is used in the .summary files.
#    Label: ...

    # Produce additional coverage / depth statistics for a set of
    # regions defined in a BED file; if no names are specified for the
    # BED records, results are named after the chromosome / contig.
#    RegionsOfInterest:
#      NAME: PATH_TO_BEDFILE
"""

_TEMPLATE_SAMPLES = """
# Targets are specified using the following structure:
#NAME_OF_TARGET:
#  NAME_OF_SAMPLE:
#    NAME_OF_LIBRARY:
#      NAME_OF_LANE: PATH_WITH_WILDCARDS
"""

_FILENAME = "SampleSheet.csv"


def build_makefile(add_full_options=True,
                   add_prefix_tmpl=True,
                   add_sample_tmpl=True):
    timestamp = datetime.datetime.now().isoformat()
    template_parts = [_TEMPLATE_TOP % (timestamp,)]

    if add_full_options:
        template_parts.append(_TEMPLATE_BAM_OPTIONS)

    if add_prefix_tmpl:
        template_parts.append(_TEMPLATE_PREFIXES)

    if add_sample_tmpl:
        template_parts.append(_TEMPLATE_SAMPLES)

    return "\n".join(template_parts)


def strip_comments(text):
    lines = text.split("\n")

    # Always include minimal header
    minimal_template = lines[:3]
    for line in lines[3:]:
        if not line.lstrip().startswith("#"):
            line = line.split("#", 1)[0].rstrip()

            # Avoid too many empty lines
            if line.strip() or minimal_template[-1].strip():
                minimal_template.append(line)

    return "\n".join(minimal_template)


def read_alignment_records(filename):
    with open(filename) as records:
        header = records.readline().strip().split(",")
        for line in records:
            yield dict(zip(header, line.strip().split(",")))


def parse_args(argv):
    parser = OptionParser("Usage: %prog [/path/to/SampleSheet.csv, ...]")
    parser.add_option("--minimal", default=False, action="store_true",
                      help="Strip comments from makefile template.")

    return parser.parse_args(argv)


def select_path(path):
    has_r1 = bool(glob.glob(path.format(Pair=1)))
    has_r2 = bool(glob.glob(path.format(Pair=2)))

    if has_r1 and not has_r2:
        # Single-ended reads
        return path.format(Pair=1)
    return path


def main(argv, pipeline="bam"):
    assert pipeline in ("bam", "trim"), pipeline

    options, paths = parse_args(argv)
    records = {}
    for root in paths:
        if os.path.isdir(root):
            filename = os.path.join(root, _FILENAME)
        else:
            root, filename = os.path.split(root)[0], root

        if not os.path.exists(filename):
            print_err("ERROR: Could not find SampleSheet file: %r" % filename)
            return 1

        for record in read_alignment_records(filename):
            libraries = records.setdefault(record["SampleID"], {})
            barcodes = libraries.setdefault(record["Index"], [])

            record["Lane"] = int(record["Lane"])
            path = "%(SampleID)s_%(Index)s_L%(Lane)03i_R{Pair}_*.fastq.gz" \
                % record
            record["Path"] = select_path(os.path.join(root, path))
            barcodes.append(record)

    template = build_makefile(add_full_options=(pipeline == "bam"))
    if options.minimal:
        template = strip_comments(template)

    print(template)

    for (sample, libraries) in records.iteritems():
        print("%s:" % sample)
        print("  %s:" % sample)
        for (library, barcodes) in libraries.iteritems():
            print("    %s:" % library)
            for record in barcodes:
                print("      {FCID}_{Lane}: {Path}".format(**record))
            print()
        print()

    if argv:
        print_info("Automatically generated makefile printed.\n"
                   "Please check for correctness before running pipeline.")
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2015 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import gzip
import string
import collections

import pysam

from paleomix.ui import \
    print_warn

from paleomix.node import \
    CommandNode, \
    Node

from paleomix.common.fileutils import \
    safe_coerce_to_frozenset

from paleomix.common.fileutils import \
    try_remove, \
    reroot_path, \
    make_dirs, \
    move_file

from paleomix.atomiccmd.command import \
    AtomicCmd

import paleomix.tools.factory as factory


class SplitReadsNode(Node):
    # TODO: Add options for kmer locations / sizes
    def __init__(self, input_files, output_dir, per_file=4000000,
                 dependencies=()):
        self._output_table = os.path.join(output_dir, "list.txt")
        self._output_files = []
        self._output_dir = output_dir
        self._per_file = per_file

        Node.__init__(self,
                      input_files=input_files,
                      output_files=(self._output_table,),
                      dependencies=dependencies)

    def _run(self, config, temp):
        Node._run(self, config, temp)

        handle = None
        for fpath in sorted(self.input_files):



            output_fpath = os.path.join(temp, "%03i.fasta.gz" % (idx,))
            with gzip.open(output_fpath, "w") as handle:
                pass # FIXME


    def _teardown(self, config, temp):
        moved_files = set()
        try:
            for fname in self._output_files:
                src_fpath = os.path.join(temp, fname)
                dst_fpath = os.path.join(self._output_dir, fname)
                move_file(src_fpath, dst_fpath)
                moved_files.add(dst_fpath)
        except:
            for fpath in moved_files:
                try_remove(fpath)
            raise




class MrCNVRPrepNode(CommandNode):
    def __init__(self, fasta_fpath, output_file, dependencies=()):
        cmd = AtomicCmd(["mrcanavar", "--prep",
                         "-fasta", "%(IN_FASTA)s",
                         "-gaps", "/dev/null",
                         "-conf", "%(OUT_CONF)s"],
                        IN_FASTA=fasta_fpath,
                        OUT_CONF=output_file)

        CommandNode.__init__(self,
                             command=cmd,
                             dependencies=dependencies)


class CatFASTANode(CommandNode):
    def __init__(self, fasta_fpaths, output_file, dependencies=()):
        cmd = factory.new("cat")
        cmd.add_multiple_values(fasta_fpaths)
        cmd.set_kwargs(OUT_STDOUT=output_file)

        CommandNode.__init__(self,
                             command=cmd.finalize(),
                             dependencies=dependencies)


class MaskBedNode(CommandNode):
    def __init__(self, fasta_fpath, bed_fpath, output_file,
                 slop=0, contig_size=float("inf"), dependencies=()):
        self._slop = slop
        self._bed_fpath = bed_fpath
        self._contig_size = contig_size

        cmd = AtomicCmd(["bedtools", "maskfasta",
                         "-fi", "%(IN_FASTA)s",
                         "-bed", "%(TEMP_OUT_BED)s",
                         "-fo", "%(OUT_FASTA)s"],
                        IN_FASTA=fasta_fpath,
                        IN_BED=bed_fpath,
                        TEMP_OUT_BED=os.path.basename(bed_fpath),
                        OUT_FASTA=output_file)

        CommandNode.__init__(self,
                             command=cmd,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)
        if self._slop:
            records = []
            with open(self._bed_fpath) as handle:
                for contig, start, end in MergeBedsNode.read_bed(self._bed_fpath):
                    start = max(0, start - self._slop)
                    end = min(self._contig_size, end + self._slop)
                    records.append((contig, start, end, None))

            with open(reroot_path(temp, self._bed_fpath), "w") as handle:
                for contig, start, end, _ in CountHitsNode.merge_beds(records):
                    handle.write("%s\t%i\t%i\n" % (contig, start, end))
        else:
            os.symlink(os.path.abspath(self._bed_fpath),
                       os.path.join(temp, os.path.basename(self._bed_fpath)))


class MergeBedsNode(Node):
    def __init__(self, hits_files, bed_files, output_file, contig,
                 max_hits=1, dependencies=()):
        self._contig = contig
        self._bed_files = {}

        for fpath in safe_coerce_to_frozenset(hits_files):
            self._bed_files[fpath] = max_hits
        for fpath in safe_coerce_to_frozenset(bed_files):
            self._bed_files[fpath] = None

        Node.__init__(self,
                      input_files=self._bed_files.keys(),
                      output_files=(output_file,),
                      dependencies=dependencies)

    def _run(self, config, temp):
        Node._run(self, config, temp)

        records = []
        contig = self._contig
        for fpath, max_score in self._bed_files.iteritems():
            bed_records = self.read_bed(fpath=fpath,
                                        contig=contig,
                                        max_score=max_score)
            for contig, start, end in bed_records:
                records.append((contig, start, end, None))

        if not records:
            # Create dummy entry
            records = [("_" + contig, 0, 1, None)]

        output_file, = self.output_files
        with open(reroot_path(temp, output_file), "w") as handle:
            for contig, start, end, _ in CountHitsNode.merge_beds(records):
                handle.write("%s\t%i\t%i\n" % (contig, start, end))

    def _teardown(self, config, temp):
        output_file, = self.output_files
        move_file(reroot_path(temp, output_file), output_file)
        Node._teardown(self, config, temp)

    @classmethod
    def read_bed(cls, fpath, contig=None, max_score=None):
        with open(fpath) as handle:
            for line in handle:
                if line.lstrip().startswith("#"):
                    continue

                fields = line.split("\t")
                if max_score is not None and fields[4] <= max_score:
                    continue

                cur_contig, start, end = fields[:3]
                if contig is None or cur_contig == contig:
                    yield cur_contig, int(start), int(end)


class CountHitsNode(Node):
    def __init__(self, input_files, output_file, dependencies=()):
        self._output_table = output_file
        Node.__init__(self,
                      input_files=input_files,
                      output_files=(output_file,),
                      dependencies=dependencies)

    def _run(self, config, temp):
        Node._run(self, config, temp)
        results = collections.defaultdict(int)
        for fpath in self.input_files:
            with gzip.open(fpath) as handle:
                for line in handle:
                    name, _ = line.split('\t', 1)
                    results[name] += 1

        selection = self.select_parse_and_merge_ranges(results)
        with open(reroot_path(temp, self._output_table), "w") as handle:
            for contig, start, end, count in selection:
                handle.write("%s\t%s\t%s\tNA\t%i\t+\n"
                             % (contig, start, end, count))

    def _teardown(self, config, temp):
        src_fpath = reroot_path(temp, self._output_table)
        move_file(src_fpath, self._output_table)
        Node._teardown(self, config, temp)

    @classmethod
    def select_parse_and_merge_ranges(cls, ranges):
        selection = []
        for key, count in ranges.iteritems():
            if count > 1:
                contig, start, end = cls.split_record(key)
                selection.append((contig, start, end, count))

        return cls.merge_beds(selection)

    @classmethod
    def split_record(cls, record):
        contig, positions = record.rsplit(":", 1)
        start, end = positions.split("-", 1)

        return contig, int(start), int(end)

    @classmethod
    def merge_beds(cls, records):
        records = list(sorted(records))
        if not records:
            return []

        merged = []
        last_contig, last_start, last_end, last_count = records[0]
        for contig, start, end, count in records:
            if last_contig != contig or last_end <= start or count != last_count:
                merged.append((last_contig, last_start, last_end, last_count))
                last_contig, last_start, last_end, last_count \
                    = contig, start, end, count
            else:
                last_end = max(end, last_end)

        merged.append((last_contig, last_start, last_end, last_count))

        return merged


class MrFastIndexNode(CommandNode):
    def __init__(self, fasta_fpath, ws=12, dependencies=()):
        cmd = AtomicCmd(["mrfast",
                         "--index", "%(TEMP_OUT_FASTA)s",
                         "--ws", str(ws)],
                        IN_FASTA=fasta_fpath,
                        TEMP_OUT_FASTA=os.path.basename(fasta_fpath),
                        OUT_INDEX=fasta_fpath + ".index")

        CommandNode.__init__(self,
                             description="<MrFastIndex: %r>" % (fasta_fpath,),
                             command=cmd,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        fasta_fpath = tuple(self.input_files)[0]
        os.symlink(os.path.abspath(fasta_fpath),
                   os.path.join(temp, os.path.basename(fasta_fpath)))


class MrFastMapNode(CommandNode):
    def __init__(self,
                 reads_fpath,
                 fasta_fpath,
                 output_prefix,
                 dependencies=()):
        cmd = AtomicCmd(["mrfast",
                         "--seqcomp", "--outcomp",
                         "--search", "%(IN_FASTA)s",
                         "--seq", "%(IN_READS)s",
                         "-o", "%(TEMP_OUT_MAPPED)s",
                         "-u", "/dev/null",
                         ],
                        IN_FASTA=fasta_fpath,
                        IN_FASTA_INDEX=fasta_fpath + ".index",
                        IN_READS=reads_fpath,
                        TEMP_OUT_MAPPED=os.path.basename(output_prefix + ".sam"),
                        OUT_MAPPED=output_prefix + ".sam.gz")

        CommandNode.__init__(self,
                             description="<MrFastMap: %r -> %r>"
                             % (reads_fpath, output_prefix),
                             command=cmd,
                             dependencies=dependencies)


class FASTAKMerNode(Node):
    def __init__(self,
                 fasta_fpath, output_dir, contigs,
                 kmer_size=36, step_size=5, per_file=4000000,
                 dependencies=()):
        self._fasta_fpath = fasta_fpath
        self._kmer_size = kmer_size
        self._step_size = step_size
        self._per_file = per_file

        n_kmers = 0
        for size in contigs.itervalues():
            n_kmers += (size - kmer_size + step_size) // step_size

        output_files = []
        for counter in xrange((n_kmers + per_file - 1) // per_file):
            output_fname = "%03i.fasta.gz" % (counter, )
            output_files.append(os.path.join(output_dir, output_fname))

        Node.__init__(self,
                      description="<FASTA Kmers: %r -> %r>" \
                          % (fasta_fpath, output_dir),
                      input_files=(fasta_fpath,),
                      output_files=output_files,
                      dependencies=dependencies)

    def _run(self, config, temp):
        Node._run(self, config, temp)

        try:
            handle = None
            records = self.fragment_fasta(self._fasta_fpath,
                                          self._kmer_size,
                                          self._step_size)
            for idx, record in enumerate(records):
                if idx % self._per_file == 0:
                    if handle:
                        handle.close()

                    fname = "%03i.fasta.gz" % (idx % self._per_file,)
                    handle = gzip.open(os.path.join(temp, fname), "w")

                handle.write(record)
        finally:
            if handle:
                handle.close()

    def _teardown(self, config, temp):
        for dst_fpath in self.output_files:
            src_path = reroot_path(temp, dst_fpath)
            move_file(src_path, dst_fpath)

    @classmethod
    def fragment_fasta(cls, filename, kmer_size, step_size):
        with open(filename) as handle:
            position = 0
            offset = 0
            cache = ""
            name = None

            while True:
                while offset + kmer_size < len(cache):
                    yield "%s:%s-%s\n%s\n" \
                        % (name, position, position + kmer_size,
                           cache[offset:offset + kmer_size])
                    offset += step_size
                    position += step_size

                line = handle.readline()
                if line.startswith(">"):
                    position = 0
                    offset = 0
                    cache = ""
                    name = line.split(None, 1)[0]
                elif line:
                    if offset:
                        cache = cache[offset:]
                        offset = 0

                    cache += line.rstrip()
                else:
                    break


class SplitFASTANode(Node):
    def __init__(self, fasta_fpath, output_dir, dependencies=()):
        self._fasta_fpath = fasta_fpath
        self._output_dir = output_dir

        self.contigs = []
        for name, size in _read_contig_list(self._fasta_fpath):
            target = _transform_seq_name(name)
            fpath = os.path.join(output_dir, target + ".fasta")
            self.contigs.append((fpath, size))

        output_files = [os.path.join(output_dir, "list.txt")]
        output_files.extend(fpath for fpath, _ in self.contigs)

        Node.__init__(self,
                      input_files=(fasta_fpath, fasta_fpath + ".fai"),
                      output_files=output_files,
                      dependencies=dependencies)

    def _run(self, _config, temp):
        try:
            last_handle = None
            with open(self._fasta_fpath) as handle:
                for line in handle:
                    if line.startswith(">"):
                        if last_handle:
                            last_handle.close()
                            last_handle = None

                        name = line.split(None, 1)[0][1:]
                        target = _transform_seq_name(name)
                        target_fpath = os.path.join(temp, target + ".fasta")

                        last_handle = open(target_fpath, "w")
                        last_handle.write(line)
                    elif last_handle:
                        last_handle.write(line)
                    else:
                        assert False, self._fasta_fpath
        finally:
            if last_handle:
                last_handle.close()

        make_dirs(self._output_dir)
        with open(os.path.join(self._output_dir, "list.txt"), "w") as handle:
            for fpath, _ in self.contigs:
                handle.write("%s\n" % (os.path.basename(fpath),))

        for fpath_dst, _ in self.contigs:
            fpath_src = os.path.join(temp, os.path.basename(fpath_dst))
            move_file(fpath_src, fpath_dst)


def _transform_seq_name(name):
    valid_chars = string.letters + string.digits
    filtered = []
    for char in name:
        if char in valid_chars:
            filtered.append(char)
        elif not filtered or filtered[-1] != "_":
            filtered.append("_")
    return "".join(filtered)


def _read_contig_list(fpath):
    fai_fpath = fpath + ".fai"
    # TODO: Check permissions / that file exists(?)

    if not os.path.exists(fai_fpath):
        print_warn("Creating index for %r; this may take a while ...\n"
                   % (fpath,))
        pysam.Fastafile(fpath).close()

    with open(fai_fpath) as handle:
        for line in handle:
            fields = line.split('\t', 2)
            yield fields[0], int(fields[1])



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from flexmock import flexmock
from nose.tools import assert_equal, assert_raises

from paleomix.common.testing import Monkeypatch

from paleomix.atomiccmd.builder import \
    AtomicCmdBuilder, \
    AtomicCmdBuilderError, \
    AtomicJavaCmdBuilder, \
    AtomicMPICmdBuilder, \
    apply_options, \
    use_customizable_cli_parameters, \
    create_customizable_cli_parameters, \
    JAVA_VERSIONS


###############################################################################
###############################################################################
# AtomicCmdBuilder: Constructor


def test_builder__simple__call():
    builder = AtomicCmdBuilder(["ls"])
    assert_equal(builder.call, ["ls"])


def test_builder__simple__str():
    builder = AtomicCmdBuilder("ls")
    assert_equal(builder.call, ["ls"])


def test_builder__simple__iterable():
    builder = AtomicCmdBuilder(iter(["ls"]))
    assert_equal(builder.call, ["ls"])


def test_builder__complex():
    builder = AtomicCmdBuilder(["java", "jar", "/a/jar"])
    assert_equal(builder.call, ["java", "jar", "/a/jar"])


def test_builder__kwargs__empty():
    builder = AtomicCmdBuilder(["ls"])
    assert_equal(builder.kwargs, {})


def test_builder__kwargs():
    expected = {"IN_FILE": "/abc/def.txt",
                "OUT_FILE": "/etc/fstab"}
    builder = AtomicCmdBuilder(["ls"], **expected)
    assert_equal(builder.kwargs, expected)


def test_builder__kwargs__set_cwd():
    builder = AtomicCmdBuilder(["ls"], set_cwd=True)
    assert_equal(builder.kwargs, {"set_cwd": True})


###############################################################################
###############################################################################
# AtomicCmdBuilder: set_option


def test_builder__set_option():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-name", "*.txt")
    assert_equal(builder.call, ["find", "-name", "*.txt"])


def test_builder__set_option__overwrite():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-name", "*.txt", fixed=False)
    builder.set_option("-name", "*.bat")
    assert_equal(builder.call, ["find", "-name", "*.bat"])


def test_builder__set_option__overwrite_fixed():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-name", "*.txt")
    assert_raises(AtomicCmdBuilderError, builder.set_option, "-name", "*.bat")


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_option

def test_builder__add_option():
    builder = AtomicCmdBuilder("find")
    builder.add_option("-name", "*.txt")
    assert_equal(builder.call, ["find", "-name", "*.txt"])


def test_builder__add_option__overwrite():
    builder = AtomicCmdBuilder("find")
    builder.add_option("-name", "*.txt")
    builder.add_option("-or")
    builder.add_option("-name", "*.bat")
    assert_equal(builder.call, ["find", "-name", "*.txt", "-or", "-name", "*.bat"])


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_option / set_option common tests

def test_builder__add_or_set_option__without_value():
    def _do_test_builder__add_or_set_option__without_value(setter):
        builder = AtomicCmdBuilder("find")
        setter(builder, "-delete")
        assert_equal(builder.call, ["find", "-delete"])
    yield _do_test_builder__add_or_set_option__without_value, AtomicCmdBuilder.add_option
    yield _do_test_builder__add_or_set_option__without_value, AtomicCmdBuilder.set_option


def test_builder__add_or_set_option__with_sep():
    def _do_test_builder__add_or_set_option__with_sep(setter):
        builder = AtomicCmdBuilder("find")
        setter(builder, "-size", "0", sep="=")
        assert_equal(builder.call, ["find", "-size=0"])
    yield _do_test_builder__add_or_set_option__with_sep, AtomicCmdBuilder.add_option
    yield _do_test_builder__add_or_set_option__with_sep, AtomicCmdBuilder.set_option


def test_builder__add_or_set_option__with_non_str_value():
    def _do_test_test_builder__add_or_set_option__with_non_str_value(setter):
        builder = AtomicCmdBuilder("find")
        setter(builder, "-size", 0)
        assert_equal(builder.call, ["find", "-size", 0])
    yield _do_test_test_builder__add_or_set_option__with_non_str_value, AtomicCmdBuilder.add_option
    yield _do_test_test_builder__add_or_set_option__with_non_str_value, AtomicCmdBuilder.set_option


def test_builder__add_or_set_option__add_and_set():
    def _do_test_builder__add_or_set_option__add_and_set(setter_1, setter_2):
        builder = AtomicCmdBuilder("find")
        setter_1(builder, "-name", "*.txt")
        assert_raises(AtomicCmdBuilderError, setter_2, builder, "-name", "*.bat")
    yield _do_test_builder__add_or_set_option__add_and_set, AtomicCmdBuilder.set_option, AtomicCmdBuilder.add_option
    yield _do_test_builder__add_or_set_option__add_and_set, AtomicCmdBuilder.add_option, AtomicCmdBuilder.set_option


def test_builder__add_or_set_option__with_non_str_key():
    def _do_test_builder__add_or_set_option__with_non_str_key(setter):
        builder = AtomicCmdBuilder("find")
        assert_raises(TypeError, setter, builder, 7913, "True")
    yield _do_test_builder__add_or_set_option__with_non_str_key, AtomicCmdBuilder.add_option
    yield _do_test_builder__add_or_set_option__with_non_str_key, AtomicCmdBuilder.set_option


def test_builder__add_or_set_option__after_finalize():
    def _do_test_builder__add_or_set_option__after_finalize(setter):
        builder = AtomicCmdBuilder("find")
        builder.finalize()
        assert_raises(AtomicCmdBuilderError, setter, builder, "-size", "1")
    yield _do_test_builder__add_or_set_option__after_finalize, AtomicCmdBuilder.add_option
    yield _do_test_builder__add_or_set_option__after_finalize, AtomicCmdBuilder.set_option


def test_builder__add_or_set_option__empty_key():
    def _do_test_builder__add_or_set_option__empty_key(setter):
        builder = AtomicCmdBuilder("find")
        assert_raises(KeyError, setter, builder, "", "1")
    yield _do_test_builder__add_or_set_option__empty_key, AtomicCmdBuilder.add_option
    yield _do_test_builder__add_or_set_option__empty_key, AtomicCmdBuilder.set_option


###############################################################################
###############################################################################
# AtomicCmdBuilder: pop_option

def test_builder__pop_option():
    def _do_test_builder__pop_option(setter):
        builder = AtomicCmdBuilder("find")
        setter(builder, "-size", "0", fixed=False)
        builder.pop_option("-size")
        assert_equal(builder.call, ["find"])
    yield _do_test_builder__pop_option, AtomicCmdBuilder.set_option
    yield _do_test_builder__pop_option, AtomicCmdBuilder.add_option


def test_builder__pop_option__last_option():
    builder = AtomicCmdBuilder("find")
    builder.add_option("-size", "0", fixed=False)
    builder.add_option("-size", "1", fixed=False)
    builder.pop_option("-size")
    assert_equal(builder.call, ["find", "-size", "0"])


def test_builder__pop_option__different_options():
    def _do_test_builder__pop_option(setter):
        builder = AtomicCmdBuilder("find")
        setter(builder, "-empty", fixed=False)
        setter(builder, "-size", "1", fixed=False)
        setter(builder, "-name", "*.txt", fixed=False)
        builder.pop_option("-size")
        assert_equal(builder.call, ["find", "-empty", "-name", "*.txt"])
    yield _do_test_builder__pop_option, AtomicCmdBuilder.set_option
    yield _do_test_builder__pop_option, AtomicCmdBuilder.add_option


def test_builder__pop_option__is_fixed():
    def _do_test_builder__pop_option__is_fixed(setter):
        builder = AtomicCmdBuilder("find")
        setter(builder, "-size", "0")
        assert_raises(AtomicCmdBuilderError, builder.pop_option, "-size")
    yield _do_test_builder__pop_option__is_fixed, AtomicCmdBuilder.set_option
    yield _do_test_builder__pop_option__is_fixed, AtomicCmdBuilder.add_option


def test_builder__pop_option__empty():
    builder = AtomicCmdBuilder("find")
    assert_raises(KeyError, builder.pop_option, "-size")


def test_builder__pop_option__missing_key():
    builder = AtomicCmdBuilder("find")
    builder.set_option("-size", 0)
    assert_raises(KeyError, builder.pop_option, "-isize")


def test_builder__pop_option__with_non_str_key():
    builder = AtomicCmdBuilder("find")
    assert_raises(TypeError, builder.pop_option, 7913)


def test_builder__pop_option__with_empty_key():
    builder = AtomicCmdBuilder("find")
    assert_raises(KeyError, builder.pop_option, "")


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_value

def test_builder__add_value():
    builder = AtomicCmdBuilder("ls")
    builder.add_value("%(IN_FILE)s")
    assert_equal(builder.call, ["ls", "%(IN_FILE)s"])


def test_builder__add_value__two_values():
    builder = AtomicCmdBuilder("ls")
    builder.add_value("%(IN_FILE)s")
    builder.add_value("%(OUT_FILE)s")
    assert_equal(builder.call, ["ls", "%(IN_FILE)s", "%(OUT_FILE)s"])


###############################################################################
###############################################################################
# AtomicCmdBuilder: set_kwargs

def test_builder__set_kwargs__called_once():
    expected = {"IN_PATH": "/a/b/", "OUT_PATH": "/dst/file"}
    builder = AtomicCmdBuilder("echo")
    builder.set_kwargs(**expected)
    assert_equal(builder.kwargs, expected)


def test_builder__set_kwargs__called_twice():
    expected = {"IN_PATH": "/a/b/", "OUT_PATH": "/dst/file"}
    builder = AtomicCmdBuilder("echo")
    builder.set_kwargs(OUT_PATH="/dst/file")
    builder.set_kwargs(IN_PATH="/a/b/")
    assert_equal(builder.kwargs, expected)


def test_builder__set_kwargs__atomiccmdbuilder():
    mock = flexmock(AtomicCmdBuilder("true"))
    mock.should_receive('finalize').and_return("finalized!")
    builder = AtomicCmdBuilder("ls", IN_BUILDER=mock)
    assert_equal(builder.kwargs, {"IN_BUILDER": "finalized!"})


def test_builder__set_kwargs__after_finalize():
    expected = {"IN_PATH": "/a/b/"}
    builder = AtomicCmdBuilder("echo")
    builder.set_kwargs(IN_PATH="/a/b/")
    builder.finalize()
    assert_raises(AtomicCmdBuilderError, builder.set_kwargs, OUT_PATH="/dst/file")
    assert_equal(builder.kwargs, expected)


def test_builder__set__kwargs__overwriting():
    expected = {"IN_PATH": "/a/b/"}
    builder = AtomicCmdBuilder("echo")
    builder.set_kwargs(IN_PATH="/a/b/")
    assert_raises(AtomicCmdBuilderError, builder.set_kwargs, IN_PATH="/dst/file")
    assert_equal(builder.kwargs, expected)


###############################################################################
###############################################################################
# AtomicCmdBuilder: finalize

def test_builder__finalized_call__simple_command():
    builder = AtomicCmdBuilder("echo")
    assert_equal(builder.finalized_call, ["echo"])


def test_builder__finalized_call__kwargs_are_instantiated():
    builder = AtomicCmdBuilder(("echo", "%(ARG1)s", "X=%(ARG2)s"),
                               ARG1="/foo/bar",
                               ARG2="zod")
    assert_equal(builder.finalized_call, ["echo", "/foo/bar", "X=zod"])


def test_builder__finalized_call__kwargs_are_instantiated__with_temp_dir():
    builder = AtomicCmdBuilder(("echo", "%(ARG)s", "%(TEMP_DIR)s"),
                               ARG="/foo/bar")
    assert_equal(builder.finalized_call, ["echo", "/foo/bar", "%(TEMP_DIR)"])


def test_builder__finalized_call__kwargs_are_instantiated__with_non_str_arg():
    builder = AtomicCmdBuilder(("echo", "%(ARG)s", 17),
                               ARG="/foo/bar")
    assert_equal(builder.finalized_call, ["echo", "/foo/bar", "17"])


###############################################################################
###############################################################################
# AtomicCmdBuilder: finalize

def test_builder__finalize__returns_singleton():
    builder = AtomicCmdBuilder("echo")
    assert builder.finalize() is builder.finalize()


def test_builder__finalize__calls_atomiccmd():
    was_called = []

    class _AtomicCmdMock:
        def __init__(self, *args, **kwargs):
            assert_equal(args, (["echo", "-out", "%(OUT_FILE)s", "%(IN_FILE)s"],))
            assert_equal(kwargs, {"IN_FILE": "/in/file",
                                  "OUT_FILE": "/out/file",
                                  "set_cwd": True})
            was_called.append(True)

    with Monkeypatch("paleomix.atomiccmd.builder.AtomicCmd", _AtomicCmdMock):
        builder = AtomicCmdBuilder("echo", set_cwd=True)
        builder.add_option("-out", "%(OUT_FILE)s")
        builder.add_value("%(IN_FILE)s")
        builder.set_kwargs(OUT_FILE="/out/file",
                           IN_FILE="/in/file")

        builder.finalize()
        assert was_called


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_multiple_options

def test_builder__add_multiple_options():
    values = ("file_a", "file_b")
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_options("-i", values)

    assert_equal(kwargs, expected)
    assert_equal(builder.kwargs, expected)
    assert_equal(builder.call, ["ls",
                                "-i", "%(IN_FILE_01)s",
                                "-i", "%(IN_FILE_02)s"])


def test_builder__add_multiple_options_with_sep():
    values = ("file_a", "file_b")
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_options("-i", values, sep="=")

    assert_equal(kwargs, expected)
    assert_equal(builder.kwargs, expected)
    assert_equal(builder.call, ["ls",
                                "-i=%(IN_FILE_01)s",
                                "-i=%(IN_FILE_02)s"])


def test_builder__add_multiple_options_with_template():
    values = ("file_a", "file_b")
    expected = {"OUT_BAM_1": "file_a", "OUT_BAM_2": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_options("-i", values, template="OUT_BAM_%i")

    assert_equal(kwargs, expected)
    assert_equal(builder.kwargs, expected)
    assert_equal(builder.call, ["ls",
                                "-i", "%(OUT_BAM_1)s",
                                "-i", "%(OUT_BAM_2)s"])


def test_builder__add_multiple_options_multiple_times():
    expected = {"IN_FILE_01": "file_a",
                "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_options("-i", ("file_a",))
    assert_equal(kwargs, {"IN_FILE_01": "file_a"})
    kwargs = builder.add_multiple_options("-i", ("file_b",))
    assert_equal(kwargs, {"IN_FILE_02": "file_b"})

    assert_equal(builder.kwargs, expected)
    assert_equal(builder.call, ["ls",
                                "-i", "%(IN_FILE_01)s",
                                "-i", "%(IN_FILE_02)s"])


###############################################################################
###############################################################################
# AtomicCmdBuilder: add_multiple_values

def test_builder__add_multiple_values():
    values = ("file_a", "file_b")
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_values(values)

    assert_equal(kwargs, expected)
    assert_equal(builder.kwargs, expected)
    assert_equal(builder.call, ["ls", "%(IN_FILE_01)s", "%(IN_FILE_02)s"])


def test_builder__add_multiple_values_with_template():
    values = ("file_a", "file_b")
    expected = {"OUT_BAM_1": "file_a", "OUT_BAM_2": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_values(values, template="OUT_BAM_%i")

    assert_equal(kwargs, expected)
    assert_equal(builder.kwargs, expected)
    assert_equal(builder.call, ["ls", "%(OUT_BAM_1)s", "%(OUT_BAM_2)s"])


def test_builder__add_multiple_values_multiple_times():
    expected = {"IN_FILE_01": "file_a", "IN_FILE_02": "file_b"}

    builder = AtomicCmdBuilder("ls")
    kwargs = builder.add_multiple_values(("file_a",))
    assert_equal(kwargs, {"IN_FILE_01": "file_a"})
    kwargs = builder.add_multiple_values(("file_b",))
    assert_equal(kwargs, {"IN_FILE_02": "file_b"})

    assert_equal(builder.kwargs, expected)
    assert_equal(builder.call, ["ls", "%(IN_FILE_01)s", "%(IN_FILE_02)s"])


###############################################################################
###############################################################################
# AtomicJavaCmdBuilder

def test_java_builder__default__no_config():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar")
    assert_equal(builder.call, ["java",
                                "-server",
                                "-Djava.io.tmpdir=%(TEMP_DIR)s",
                                "-Djava.awt.headless=true",
                                "-XX:+UseSerialGC",
                                "-Xmx4g",
                                "-jar", "%(AUX_JAR)s"])


def test_java_builder__defaults__call():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar", temp_root="/disk/tmp")
    assert_equal(builder.call, ["java",
                                "-server",
                                "-Djava.io.tmpdir=/disk/tmp",
                                "-Djava.awt.headless=true",
                                "-XX:+UseSerialGC",
                                "-Xmx4g",
                                "-jar", "%(AUX_JAR)s"])


def test_java_builder__defaults__kwargs():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar")
    assert_equal(builder.kwargs, {"AUX_JAR": "/path/Foo.jar",
                                  "CHECK_JRE": JAVA_VERSIONS[(1, 6)]})


def test_java_builder__multithreaded_gc():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar",
                                   temp_root="/disk/tmp",
                                   gc_threads=3)
    assert_equal(builder.call, ["java",
                                "-server",
                                "-Djava.io.tmpdir=/disk/tmp",
                                "-Djava.awt.headless=true",
                                "-XX:ParallelGCThreads=3",
                                "-Xmx4g",
                                "-jar", "%(AUX_JAR)s"])


def test_java_builder__multithreaded_gc__zero_or_negative_threads():
    assert_raises(ValueError, AtomicJavaCmdBuilder, "/path/Foo.jar", gc_threads=0)
    assert_raises(ValueError, AtomicJavaCmdBuilder, "/path/Foo.jar", gc_threads=-1)


def test_java_builder__multithreaded_gc__non_int_threads():
    assert_raises(TypeError, AtomicJavaCmdBuilder, "/path/Foo.jar", gc_threads="3")


def test_java_builder__kwargs():
    builder = AtomicJavaCmdBuilder("/path/Foo.jar", set_cwd=True)
    assert_equal(builder.kwargs, {"AUX_JAR": "/path/Foo.jar",
                                  "set_cwd": True,
                                  "CHECK_JRE": JAVA_VERSIONS[(1, 6)]})


###############################################################################
###############################################################################
# AtomicMPICmdBuilder

def test_mpi_builder__defaults__str():
    builder = AtomicMPICmdBuilder("ls")
    assert_equal(builder.call, ["ls"])
    assert_equal(builder.kwargs, {"EXEC_MPI": "mpirun"})


def test_mpi_builder__multithreaded__str():
    builder = AtomicMPICmdBuilder("ls", threads=3)
    assert_equal(builder.call, ["mpirun", "-n", 3, "ls"])
    assert_equal(builder.kwargs, {"EXEC_MAIN": "ls"})


def test_mpi_builder__defaults__complex_cmd():
    builder = AtomicMPICmdBuilder(["python", "/foo/run.py"])
    assert_equal(builder.call, ["python", "/foo/run.py"])
    assert_equal(builder.kwargs, {"EXEC_MPI": "mpirun"})


def test_mpi_builder__multithreaded__complex_cmd():
    builder = AtomicMPICmdBuilder(["python", "/foo/run.py"], threads=3)
    assert_equal(builder.call, ["mpirun", "-n", 3, "python", "/foo/run.py"])
    assert_equal(builder.kwargs, {"EXEC_MAIN": "python"})


def test_mpi_builder__kwargs():
    builder = AtomicMPICmdBuilder("ls", set_cwd=True)
    assert_equal(builder.kwargs, {"set_cwd": True, "EXEC_MPI": "mpirun"})


def test_mpi_builder__threads__zero_or_negative():
    assert_raises(ValueError, AtomicMPICmdBuilder, "ls", threads=0)
    assert_raises(ValueError, AtomicMPICmdBuilder, "ls", threads=-1)


def test_mpi_builder__threads__non_int():
    assert_raises(TypeError, AtomicMPICmdBuilder, "ls", threads="3")


###############################################################################
###############################################################################
# create_customizable_cli_parameters

def test_custom_cli__single_named_arg():
    class SingleNamedArg:
        @create_customizable_cli_parameters
        def customize(cls, argument):
            return {}

    value = "A value"
    obj = SingleNamedArg.customize(value)
    assert_equal(obj.argument, value)


def test_custom_cli__adding_new_values():
    class SingleNamedArg:
        @create_customizable_cli_parameters
        def customize(cls):
            return {"dynamic": 12345}

    obj = SingleNamedArg.customize()
    assert_equal(obj.dynamic, 12345)


def test_custom_cli__multiple_named_args():
    class SingleNamedArg:
        @create_customizable_cli_parameters
        def customize(cls, first, second):
            return {}

    obj = SingleNamedArg.customize(123, 456)
    assert_equal(obj.first,  123)
    assert_equal(obj.second, 456)


def test_custom_cli__only_customize_is_valid_function_name():
    try:
        class ClassWithMisnamedFunction:
            @create_customizable_cli_parameters
            def not_called_customize(cls, first, second):
                return {}  # pragma: no coverage

        assert False, "ValueError not raised"  # pragma: no coverage
    except ValueError:
        pass


###############################################################################
###############################################################################
# use_customizable_cli_parameters


###############################################################################
###############################################################################
# apply_options

def test_apply_options__single_option__default_pred__set_when_pred_is_true():
    mock = flexmock()
    mock.should_receive('set_option').with_args('--foo', 17).once()
    apply_options(mock, {"--foo": 17})


def test_apply_options__single_option__default_pred__ignore_when_pred_is_false():
    mock = flexmock()
    apply_options(mock, {"Other": None})


def _user_pred(key):
    return key.startswith("FOO")


def test_apply_options__single_option__user_pred__set_when_pred_is_true():
    mock = flexmock()
    mock.should_receive('set_option').with_args('FOO_BAR', 17).once()
    apply_options(mock, {"FOO_BAR": 17}, _user_pred)


def test_apply_options__single_option__user_pred__ignore_when_pred_is_false():
    mock = flexmock()
    apply_options(mock, {"BAR_FOO": 17}, _user_pred)


def test_apply_options__single_option__boolean__set_when_value_is_true():
    mock = flexmock()
    mock.should_receive('set_option').with_args('-v')
    apply_options(mock, {"-v": True})


def test_apply_options__single_option__boolean__set_when_value_is_none():
    mock = flexmock()
    mock.should_receive('set_option').with_args('-v')
    apply_options(mock, {"-v": None})


def test_apply_options__single_option__boolean__pop_when_value_is_false():
    mock = flexmock()
    mock.should_receive('pop_option').with_args('-v')
    apply_options(mock, {"-v": False})


def test_apply_options__multiple_option():
    mock = flexmock()
    mock.should_receive('add_option').with_args('--foo', 3).once()
    mock.should_receive('add_option').with_args('--foo', 17).once()
    apply_options(mock, {"--foo": [3, 17]})


def test_apply_options__boolean_and_none_is_single_value_only():
    mock = flexmock()
    assert_raises(TypeError, apply_options, mock, {"--foo": [True]})
    assert_raises(TypeError, apply_options, mock, {"--foo": [False]})
    assert_raises(TypeError, apply_options, mock, {"--foo": [None]})


def test_apply_options__unexpected_types_in_values():
    mock = flexmock()
    assert_raises(TypeError, apply_options, mock, {"--foo": object()})
    assert_raises(TypeError, apply_options, mock, {"--foo": iter([])})
    assert_raises(TypeError, apply_options, mock, {"--foo": {}})
    assert_raises(TypeError, apply_options, mock, {"--foo": set()})


def test_apply_options__non_string_types_in_keys():
    mock = flexmock()
    assert_raises(TypeError, apply_options, mock, {1: 17})
    assert_raises(TypeError, apply_options, mock, {("foo",): 17})


def test_apply_options__not_dict_like():
    mock = flexmock()
    assert_raises(TypeError, apply_options, mock, None)
    assert_raises(TypeError, apply_options, mock, [1, 2, 3])



#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import optparse

import paleomix
import paleomix.ui

from paleomix.config import \
     ConfigError, \
     PerHostValue, \
     PerHostConfig, \
     migrate_config


def _run_config_parser(argv, pipeline_variant):
    per_host_cfg = PerHostConfig("bam_pipeline")
    pipeline_variant = "%s_pipeline" % (pipeline_variant,)

    usage_str = "paleomix %s <command> [options] [makefiles]" % (pipeline_variant,)
    version_str = "paleomix %s %s" % (pipeline_variant, paleomix.__version__)
    parser = optparse.OptionParser(usage=usage_str, version=version_str)

    paleomix.ui.add_optiongroup(parser,
                                ui_default=PerHostValue("running"),
                                color_default=PerHostValue("on"))
    paleomix.logger.add_optiongroup(parser, default = PerHostValue("warning"))

    group = optparse.OptionGroup(parser, "Scheduling")
    group.add_option("--dry-run", action="store_true", default=False,
                     help="If passed, only a dry-run in performed, the "
                          "dependency tree is printed, and no tasks are "
                          "executed.")
    group.add_option("--max-threads", type=int, default=per_host_cfg.max_threads,
                     help="Maximum number of threads to use in total [%default]")
    group.add_option("--adapterremoval-max-threads", type=int, default=PerHostValue(1),
                     help="Maximum number of threads to use per AdapterRemoval instance [%default]")
    group.add_option("--bowtie2-max-threads", type=int, default=PerHostValue(1),
                     help="Maximum number of threads to use per Bowtie2 instance [%default]")
    group.add_option("--bwa-max-threads", type=int, default=PerHostValue(1),
                     help="Maximum number of threads to use per BWA instance [%default]")
    group.add_option("--gatk-max-threads", type=int, default=PerHostValue(1),
                     help="Maximum number of threads to use per GATK instance [%default]")
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, "Required paths")
    group.add_option("--jar-root", default = PerHostValue("~/install/jar_root", is_path = True),
                     help = "Folder containing Picard JARs (http://picard.sf.net), " \
                            "and GATK (www.broadinstitute.org/gatk). " \
                            "The latter is only required if realigning is enabled. " \
                            "[%default]")
    group.add_option("--temp-root", default = per_host_cfg.temp_root,
                     help = "Location for temporary files and folders [%default/]")
    group.add_option("--destination", default = None,
                     help = "The destination folder for result files. By default, files will be "
                            "placed in the same folder as the makefile which generated it.")
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, "Files and executables")
    group.add_option("--list-output-files", action = "store_true", default = False,
                     help = "List all files generated by pipeline for the makefile(s).")
    group.add_option("--list-executables", action="store_true", default=False,
                     help="List all executables required by the pipeline, "
                          "with version requirements (if any).")
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, "Misc")
    group.add_option("--jre-option", dest = "jre_options", action = "append", default = PerHostValue([]),
                     help = "May be specified one or more times with options to be passed "
                            "tot the JRE (Jave Runtime Environment); e.g. to change the "
                            "maximum amount of memory (default is -Xmx4g)")
    group.add_option("--to-dot-file", dest="dot_file",
                     help="Write dependency tree to the specified dot-file.")
    parser.add_option_group(group)

    return per_host_cfg.parse_args(parser, argv)


def parse_config(argv, pipeline_variant):
    migrate_config()

    config, args = _run_config_parser(argv, pipeline_variant)
    paleomix.ui.set_ui_colors(config.ui_colors)

    return config, args




__all__ = ['Mark', 'YAMLError', 'MarkedYAMLError']

class Mark(object):

    def __init__(self, name, index, line, column, buffer, pointer):
        self.name = name
        self.index = index
        self.line = line
        self.column = column
        self.buffer = buffer
        self.pointer = pointer

    def get_snippet(self, indent=4, max_length=75):
        if self.buffer is None:
            return None
        head = ''
        start = self.pointer
        while start > 0 and self.buffer[start-1] not in u'\0\r\n\x85\u2028\u2029':
            start -= 1
            if self.pointer-start > max_length//2-1:
                head = ' ... '
                start += 5
                break
        tail = ''
        end = self.pointer
        while end < len(self.buffer) and self.buffer[end] not in u'\0\r\n\x85\u2028\u2029':
            end += 1
            if end-self.pointer > max_length//2-1:
                tail = ' ... '
                end -= 5
                break
        snippet = self.buffer[start:end].encode('utf-8')
        return ' '*indent + head + snippet + tail + '\n'  \
                + ' '*(indent+self.pointer-start+len(head)) + '^'

    def __str__(self):
        snippet = self.get_snippet()
        where = "  in \"%s\", line %d, column %d"   \
                % (self.name, self.line+1, self.column+1)
        if snippet is not None:
            where += ":\n"+snippet
        return where

class YAMLError(Exception):
    pass

class MarkedYAMLError(YAMLError):

    def __init__(self, context=None, context_mark=None,
            problem=None, problem_mark=None, note=None):
        self.context = context
        self.context_mark = context_mark
        self.problem = problem
        self.problem_mark = problem_mark
        self.note = note

    def __str__(self):
        lines = []
        if self.context is not None:
            lines.append(self.context)
        if self.context_mark is not None  \
            and (self.problem is None or self.problem_mark is None
                    or self.context_mark.name != self.problem_mark.name
                    or self.context_mark.line != self.problem_mark.line
                    or self.context_mark.column != self.problem_mark.column):
            lines.append(str(self.context_mark))
        if self.problem is not None:
            lines.append(self.problem)
        if self.problem_mark is not None:
            lines.append(str(self.problem_mark))
        if self.note is not None:
            lines.append(self.note)
        return '\n'.join(lines)




#!/usr/bin/python
#
# Copyright (c) 2016 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix.common.rtools as rtools
import paleomix.tools.factory as factory

from paleomix.atomiccmd.command import AtomicCmd
from paleomix.common.formats.newick import Newick
from paleomix.node import CommandNode

from paleomix.tools.zonkey.common import RSCRIPT_VERSION


class MitoConsensusNode(CommandNode):
    def __init__(self, database, bamfile, output_prefix, dependencies=()):
        cmd = factory.new("build_mito")
        cmd.add_value("%(IN_DATABASE)s")
        cmd.add_value("%(IN_BAMFILE)s")
        cmd.add_value("%(TEMP_OUT_PREFIX)s")

        cmd.set_kwargs(IN_DATABASE=database,
                       IN_BAMFILE=bamfile,
                       TEMP_OUT_PREFIX=os.path.basename(output_prefix),
                       OUT_PHYLIP=output_prefix + ".phy",
                       OUT_FASTA=output_prefix + ".fasta",
                       OUT_SUMMARY=output_prefix + ".summary")

        CommandNode.__init__(self,
                             description="<MitoConsensus -> '%s.*'>"
                             % (output_prefix,),
                             command=cmd.finalize(),
                             dependencies=dependencies)


class DrawPhylogenyNode(CommandNode):
    def __init__(self, samples, treefile, bootstraps, output_prefix,
                 dependencies=()):
        rscript = rtools.rscript("zonkey", "tinytree.r")

        cmd = AtomicCmd(("Rscript", rscript,
                         "%(TEMP_OUT_FILE)s",
                         "%(IN_SAMPLES)s",
                         "%(TEMP_OUT_PREFIX)s"),
                        AUX_RSCRIPT=rscript,
                        IN_SAMPLES=samples,
                        IN_FILE=treefile,
                        IN_BOOTSTRAPS=bootstraps,
                        TEMP_OUT_FILE="rerooted.newick",
                        TEMP_OUT_PREFIX=os.path.basename(output_prefix),
                        OUT_TREE_PDF=output_prefix + ".pdf",
                        OUT_TREE_PNG=output_prefix + ".png",
                        CHECK_RSCRIPT=RSCRIPT_VERSION,
                        CHECK_RSCRIPT_APE=rtools.requirement("ape"),
                        CHECK_RSCRIPT_GGPLOT2=rtools.requirement("ggplot2"),
                        CHECK_RSCRIPT_GRID=rtools.requirement("grid"))

        self._treefile = treefile
        self._bootstraps = bootstraps

        CommandNode.__init__(self,
                             description="<DrawPhylogeny -> '%s.*'>"
                             % (output_prefix,),
                             command=cmd,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        with open(self._bootstraps) as handle:
            bootstraps = [Newick.from_string(line.strip())
                          for line in handle]

        with open(self._treefile) as handle:
            tree = Newick.from_string(handle.read().strip())

        tree = tree.reroot_on_midpoint()
        tree = tree.add_support(bootstraps, "{Percentage}")
        with open(os.path.join(temp, "rerooted.newick"), "w") as handle:
            handle.write("{}\n".format(tree))

        CommandNode._setup(self, config, temp)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from paleomix.node import \
    CommandNode, \
    NodeError
from paleomix.atomiccmd.builder import \
    AtomicCmdBuilder, \
    use_customizable_cli_parameters, \
    create_customizable_cli_parameters
from paleomix.common.fileutils import \
    reroot_path
from paleomix.common.formats.msa import \
    MSA, \
    MSAError
import paleomix.common.versions as versions


MAFFT_VERSION = versions.Requirement(call   = ("mafft", "--version"),
                                     search = r"v(\d+)\.(\d+)",
                                     checks = versions.GE(7, 0))


# Presets mainly taken from
# http://mafft.cbrc.jp/alignment/software/algorithms/algorithms.html
_PRESETS = {
    "mafft"    : ["mafft"],
    "auto"     : ["mafft", "--auto"],
    "fft-ns-1" : ["mafft-fftns", "--retree", 1],
    "fft-ns-2" : ["mafft-fftns"],
    "fft-ns-i" : ["mafft-fftnsi"],
    "nw-ns-i"  : ["mafft-nwnsi"],
    "l-ins-i"  : ["mafft-linsi"],
    "e-ins-i"  : ["mafft-einsi"],
    "g-ins-i"  : ["mafft-ginsi"],
    }




class MAFFTNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_file, output_file, algorithm = "auto", dependencies = ()):
        command = AtomicCmdBuilder(_PRESETS[algorithm.lower()])
        command.add_value("%(IN_FASTA)s")
        command.set_kwargs(IN_FASTA   = input_file,
                           OUT_STDOUT = output_file,
                           CHECK_VERSION = MAFFT_VERSION)

        return {"command"      : command,
                "dependencies" : dependencies}


    @use_customizable_cli_parameters
    def __init__(self, parameters):
        self._output_file = parameters.output_file
        description = "<MAFFTNode (%s): '%s' -> '%s'>" \
                % (parameters.algorithm,
                   parameters.input_file,
                   parameters.output_file)

        CommandNode.__init__(self,
                             command      = parameters.command.finalize(),
                             description  = description,
                             dependencies = parameters.dependencies)

    def _teardown(self, config, temp):
        # Validate output from MAFFT
        output_file = reroot_path(temp, self._output_file)
        try:
            MSA.from_file(output_file)
        except MSAError, error:
            raise NodeError("Invalid MSA produced by MAFFT:\n%s" % (error,))
        CommandNode._teardown(self, config, temp)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import CommandNode, NodeError
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.builder import \
     AtomicCmdBuilder, \
     use_customizable_cli_parameters, \
     create_customizable_cli_parameters

from paleomix.atomiccmd.sets import ParallelCmds
from paleomix.nodes.samtools import SAMTOOLS_VERSION
from paleomix.common.fileutils import \
     describe_paired_files, \
     missing_files

import paleomix.common.versions as versions
import paleomix.tools.factory as factory


BWA_VERSION = versions.Requirement(call=("bwa",),
                                   search=r"Version: (\d+)\.(\d+)\.(\d+)",
                                   checks=versions.Or(versions.EQ(0, 5, 9),
                                                      versions.EQ(0, 5, 10),
                                                      versions.EQ(0, 6, 2),
                                                      versions.GE(0, 7, 9)))

BWA_VERSION_07x = versions.Requirement(call=("bwa",),
                                       search=r"Version: (\d+)\.(\d+)\.(\d+)",
                                       checks=versions.GE(0, 7, 9))


class BWAIndexNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_file, prefix=None, dependencies=()):
        prefix = prefix if prefix else input_file
        params = _get_bwa_template(("bwa", "index"), prefix, iotype="OUT",
                                   IN_FILE=input_file,
                                   TEMP_OUT_PREFIX=os.path.basename(prefix),
                                   CHECK_BWA=BWA_VERSION)

        # Input fasta sequence
        params.add_value("%(IN_FILE)s")
        # Destination prefix, in temp folder
        params.set_option("-p", "%(TEMP_OUT_PREFIX)s")

        return {"prefix": prefix,
                "command": params,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = parameters.command.finalize()
        description = "<BWA Index '%s' -> '%s.*'>" % (parameters.input_file,
                                                      parameters.prefix)
        CommandNode.__init__(self,
                             command=command,
                             description=description,
                             dependencies=parameters.dependencies)


class BWABacktrack(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_file, output_file, reference, prefix, threads=1,
                  dependencies=()):
        _check_bwa_prefix(reference)
        threads = _get_max_threads(reference, threads)

        aln_in = _build_cat_command(input_file, "uncompressed_input")
        aln = _get_bwa_template(("bwa", "aln"), prefix,
                                TEMP_IN_FILE="uncompressed_input",
                                OUT_STDOUT=output_file,
                                CHECK_BWA=BWA_VERSION)
        aln.add_value(prefix)
        aln.add_value("%(TEMP_IN_FILE)s")
        aln.set_option("-t", threads)

        return {"commands": {"aln_in": aln_in, "aln": aln},
                "order": ["aln_in", "aln"],
                "threads": threads,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = ParallelCmds([parameters.commands[key].finalize()
                               for key in parameters.order])

        description \
            = _get_node_description(name="BWA Backtrack",
                                    algorithm="SE",
                                    input_files_1=(parameters.input_file,),
                                    input_files_2=(),
                                    prefix=parameters.prefix,
                                    threads=parameters.threads)

        CommandNode.__init__(self,
                             command=command,
                             description=description,
                             threads=parameters.threads,
                             dependencies=parameters.dependencies)

    def _setup(self, _config, temp):
        os.mkfifo(os.path.join(temp, "uncompressed_input"))


class BWASamse(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_file_fq, input_file_sai, output_file,
                  reference, prefix, dependencies=()):
        _check_bwa_prefix(reference)

        samse_in = _build_cat_command(input_file_fq, "uncompressed_input")
        samse = _get_bwa_template(("bwa", "samse"), prefix,
                                  IN_FILE_SAI=input_file_sai,
                                  TEMP_IN_FQ="uncompressed_input",
                                  OUT_STDOUT=AtomicCmd.PIPE,
                                  CHECK_BWA=BWA_VERSION)
        samse.add_value(prefix)
        samse.add_value("%(IN_FILE_SAI)s")
        samse.add_value("%(TEMP_IN_FQ)s")

        order, commands = _process_output(samse, output_file, reference)
        commands["sam_in"] = samse_in
        commands["sam"] = samse

        return {"commands": commands,
                "order": ["sam_in", "sam"] + order,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = ParallelCmds([parameters.commands[key].finalize()
                               for key in parameters.order])

        input_file = parameters.input_file_fq
        description = _get_node_description(name="BWA Samse",
                                            algorithm="SE",
                                            input_files_1=(input_file,),
                                            input_files_2=(),
                                            prefix=parameters.prefix)

        CommandNode.__init__(self,
                             command=command,
                             description=description,
                             dependencies=parameters.dependencies)

    def _setup(self, _config, temp):
        os.mkfifo(os.path.join(temp, "uncompressed_input"))


class BWASampe(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls,
                  input_file_fq_1, input_file_fq_2,
                  input_file_sai_1, input_file_sai_2,
                  output_file, reference, prefix, dependencies=()):
        _check_bwa_prefix(reference)

        sampe_in_1 = _build_cat_command(input_file_fq_1,
                                        "uncompressed_input_1")
        sampe_in_2 = _build_cat_command(input_file_fq_2,
                                        "uncompressed_input_2")

        sampe = _get_bwa_template(("bwa", "sampe"), prefix,
                                  IN_FILE_SAI_1=input_file_sai_1,
                                  IN_FILE_SAI_2=input_file_sai_2,
                                  TEMP_IN_FQ_1="uncompressed_input_1",
                                  TEMP_IN_FQ_2="uncompressed_input_2",
                                  OUT_STDOUT=AtomicCmd.PIPE,
                                  CHECK_BWA=BWA_VERSION)
        sampe.add_value(prefix)
        sampe.add_value("%(IN_FILE_SAI_1)s")
        sampe.add_value("%(IN_FILE_SAI_2)s")
        sampe.add_value("%(TEMP_IN_FQ_1)s")
        sampe.add_value("%(TEMP_IN_FQ_2)s")

        order, commands = _process_output(sampe, output_file, reference)
        commands["sam_in_1"] = sampe_in_1
        commands["sam_in_2"] = sampe_in_2
        commands["sam"] = sampe

        return {"commands": commands,
                "order": ["sam_in_1", "sam_in_2", "sam"] + order,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        command = ParallelCmds([parameters.commands[key].finalize()
                               for key in parameters.order])

        input_file_1 = parameters.input_file_fq_1
        input_file_2 = parameters.input_file_fq_2
        description = _get_node_description(name="BWA Samse",
                                            algorithm="SE",
                                            input_files_1=(input_file_1,),
                                            input_files_2=(input_file_2,),
                                            prefix=parameters.prefix)

        CommandNode.__init__(self,
                             command=command,
                             description=description,
                             dependencies=parameters.dependencies)

    def _setup(self, _config, temp):
        os.mkfifo(os.path.join(temp, "uncompressed_input_1"))
        os.mkfifo(os.path.join(temp, "uncompressed_input_2"))


class BWAAlgorithmNode(CommandNode):
    @create_customizable_cli_parameters
    def customize(cls, input_file_1, output_file, reference, prefix,
                  input_file_2=None, threads=1, algorithm="mem",
                  dependencies=()):
        if algorithm not in ("mem", "bwasw"):
            raise NotImplementedError("BWA algorithm %r not implemented"
                                      % (algorithm,))

        threads = _get_max_threads(reference, threads)

        zcat_1 = _build_cat_command(input_file_1, "uncompressed_input_1")
        aln = _get_bwa_template(("bwa", algorithm), prefix,
                                TEMP_IN_FILE_1="uncompressed_input_1",
                                OUT_STDOUT=AtomicCmd.PIPE,
                                CHECK_BWA=BWA_VERSION_07x)
        aln.add_value(prefix)
        aln.add_value("%(TEMP_IN_FILE_1)s")

        _, commands = _process_output(aln, output_file, reference)
        commands["aln"] = aln
        commands["zcat_1"] = zcat_1
        if input_file_2:
            aln.add_value("%(TEMP_IN_FILE_2)s")
            aln.set_kwargs(**{"TEMP_IN_FILE_2": "uncompressed_input_2"})
            zcat_2 = _build_cat_command(input_file_2, "uncompressed_input_2")
            commands["zcat_2"] = zcat_2
        else:
            # Ensure that the pipe is automatically removed
            aln.set_kwargs(**{"TEMP_OUT_FILE_2": "uncompressed_input_2"})

        aln.set_option("-t", threads)
        # Mark alternative hits as secondary; required by e.g. Picard
        aln.set_option("-M")

        commands["aln"] = aln
        return {"commands": commands,
                "threads": threads,
                "dependencies": dependencies}

    @use_customizable_cli_parameters
    def __init__(self, parameters):
        _check_bwa_prefix(parameters.prefix)
        algorithm = parameters.algorithm.upper()
        algorithm += "_PE" if parameters.input_file_2 else "_SE"
        description = _get_node_description(name="BWA",
                                            algorithm=algorithm,
                                            input_files_1=parameters.input_file_1,
                                            input_files_2=parameters.input_file_2,
                                            prefix=parameters.prefix)

        command = ParallelCmds([cmd.finalize()
                                for cmd in parameters.commands.itervalues()])
        CommandNode.__init__(self,
                             command=command,
                             description=description,
                             threads=parameters.threads,
                             dependencies=parameters.dependencies)

    def _setup(self, _config, temp):
        os.mkfifo(os.path.join(temp, "uncompressed_input_1"))
        os.mkfifo(os.path.join(temp, "uncompressed_input_2"))


def _process_output(stdin, output_file, reference, run_fixmate=False):
    convert = factory.new("cleanup")
    convert.set_option("--fasta", "%(IN_FASTA_REF)s")
    convert.set_option("--temp-prefix", "%(TEMP_OUT_PREFIX)s")
    convert.set_kwargs(IN_STDIN=stdin,
                       IN_FASTA_REF=reference,
                       OUT_STDOUT=output_file,
                       TEMP_OUT_PREFIX="bam_cleanup",
                       CHECK_SAMTOOLS=SAMTOOLS_VERSION)

    if run_fixmate:
        convert.set_option('--paired-ended')

    try:
        if SAMTOOLS_VERSION.version >= (1,):
            convert.set_option('--samtools1x', 'yes')
        else:
            convert.set_option('--samtools1x', 'no')
    except versions.VersionRequirementError:
        pass

    return ["convert"], {"convert": convert}


def _get_bwa_template(call, prefix, iotype="IN", **kwargs):
    extensions = ["amb", "ann", "bwt", "pac", "sa"]
    try:
        if BWA_VERSION.version < (0, 6, 0):
            extensions.extend(("rbwt", "rpac", "rsa"))
    except versions.VersionRequirementError:
        pass  # Ignored here, handled elsewhere

    params = AtomicCmdBuilder(call, **kwargs)
    for postfix in extensions:
        key = "%s_PREFIX_%s" % (iotype, postfix.upper())
        params.set_kwargs(**{key: (prefix + "." + postfix)})

    return params


def _get_max_threads(reference, threads):
    """Returns the maximum number of threads to use when mapping against a
    given reference sequence. This is done since very little gain is obtained
    when using multiple threads for a small genome (e.g. < 1MB). If the
    reference falls below this size, only 1 thread is used (returned),
    otherwise the requested number of threads is returned.
    """
    if reference not in _PREFIX_SIZE_CACHE:
        if not os.path.exists(reference):
            _PREFIX_SIZE_CACHE[reference] = None
        else:
            _PREFIX_SIZE_CACHE[reference] = os.path.getsize(reference)

    prefix_size = _PREFIX_SIZE_CACHE[reference]
    if prefix_size is None or prefix_size >= 2 ** 20:  # > 1 MB
        return threads
    return 1
_PREFIX_SIZE_CACHE = {}


def _check_bwa_prefix(prefix):
    """Checks that a given prefix is compatible with the currently
    installed version of BWA. This is required in order to allow
    auto-indexing of prefixes, as indexes produced by v0.5.x and
    by 0.6+ are not only incompatible, but differs in the files
    produced, with 0.5.x producing a handful of additional files.

    As a consequence, simply using normal input-file dependencies
    would result in prefixes being re-indexed if the version of
    BWA was changed from 0.6+ to 0.5.x, and in failures during
    runtime if the version was changed from 0.5.x to 0.6+.

    This function treats that a difference in the version of BWA
    installed and the version implied by the prefix files is an
    error, and therefore requires user intervention."""
    if prefix in _PREFIXES_CHECKED:
        return
    _PREFIXES_CHECKED.add(prefix)

    try:
        bwa_version = BWA_VERSION.version
    except versions.VersionRequirementError:
        return  # Ignored here, reported elsewhere

    # Files unique to v0.5.x
    v05x_files = set((prefix + ext) for ext in (".rbwt", ".rpac", ".rsa"))
    # Files common to v0.5.x, v0.6.x, and v0.7.x
    common_files = set((prefix + ext)
                       for ext in (".amb", ".ann", ".bwt", ".pac", ".sa"))
    all_files = v05x_files | common_files
    current_files = all_files - set(missing_files(all_files))

    expected_version = None
    if (current_files & common_files):
        if bwa_version >= (0, 6, 0):
            if (current_files & v05x_files):
                expected_version = "v0.5.x"
        elif bwa_version < (0, 6, 0):
            if not (current_files & v05x_files):
                expected_version = "v0.6.x or later"

    if expected_version:
        raise NodeError("BWA version is v%s, but prefix appears to be created using %s!\n"
                        "  Your copy of BWA may have changed, or you may be using the wrong\n"
                        "  prefix. To resolve this issue, either change your prefix, re-install\n"
                        "  BWA %s, or remove the prefix files at\n"
                        "    $ ls %s.*" \
                        % (".".join(map(str, bwa_version)), expected_version, expected_version, prefix))
_PREFIXES_CHECKED = set()


def _build_cat_command(input_file, output_file):
    cat = factory.new("cat")
    cat.set_option("--output", "%(TEMP_OUT_CAT)s")
    cat.add_value("%(IN_ARCHIVE)s")
    cat.set_kwargs(TEMP_OUT_CAT=output_file,
                   IN_ARCHIVE=input_file)
    return cat


def _get_node_description(name, algorithm, input_files_1, input_files_2 = (), prefix = None, threads = 1):
    threads_str     = ("%i threads" % (threads,)) if (threads > 1) else None
    prefix_str      = os.path.basename(prefix) if prefix else None
    if prefix_str.endswith(".fasta") or prefix_str.endswith(".fa"):
        prefix_str  = prefix_str.rsplit(".", 1)[0]
    file_desc       = describe_paired_files(input_files_1, input_files_2 or ())
    info            = ", ".join(filter(None, (prefix_str, algorithm, threads_str)))

    return "<%s (%s): %s>" % (name, info, file_desc)



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import with_statement

import sys
import optparse
import collections

import pysam

import paleomix.common.vcfwrap as vcfwrap


_INF = float("inf")
# Rough number of records to keep in memory at once
_CHUNK_SIZE = 10000


def add_varfilter_options(parser):
    group = optparse.OptionGroup(parser, "varFilter: Novel options")
    group.add_option("--homozygous-chromosome", action="append", default=[],
                     help="Filter heterozygous SNPs observed on this "
                          "chromosome (e.g. chrX) %default.")
    group.add_option("-q", "--min-quality", type=int, default=30,
                     help="Minimum Phred score recorded in the QUAL column "
                          "[%default]")
    group.add_option("-f", "--min-allele-frequency", type=float, default=0.2,
                     help="Minimum frequency of the alleles at heterozygous "
                          "sites [%default]. WARNING: A pileup must be "
                          "provided for multi-allelic sites to be filtered!")
    group.add_option("-b", "--pileup", default=None,
                     help="Tabix indexed pileup for multi-allelic sites. This "
                          "is required for such sites to be filtered using "
                          "the --min-allele-frequency filter.")
    group.add_option("-k", "--keep-ambigious-genotypes",
                     default=False, action="store_true",
                     help="Keep SNPs without a most likely genotype "
                          "(based on PL) [%default]")
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, "varFilter: Derived options")
    # Options adapted from varFilter
    group.add_option("-Q", "--min-mapping-quality", type=int, default=10,
                     help="Minimum RMS mapping quality for SNPs [%default]")
    group.add_option("-d", "--min-read-depth", type=int, default=8,
                     help="Minimum read depth [%default]")
    group.add_option("-D", "--max-read-depth", type=int, default=10000000,
                     help="Maximum read depth [%default]")
    group.add_option("-a", "--min-num-alt-bases", type=int, default=2,
                     help="Minimum number of alternative bases observed for "
                          "variants [%default]")
    group.add_option("-w", "--min-distance-to-indels", type=int, default=3,
                     help="SNP within INT bp around a gap to be filtered "
                          "[%default]")
    group.add_option("-W", "--min-distance-between-indels",
                     type=int, default=10,
                     help="Window size for filtering adjacent gaps "
                          "[%default]")
    group.add_option("-1", "--min-strand-bias", type=float, default=1e-4,
                     help="Min P-value for strand bias (given PV4) "
                          "[%default]")
    group.add_option("-2", "--min-baseq-bias", type=float, default=1e-100,
                     help="Min P-value for baseQ bias (given PV4) "
                          "[%default]")
    group.add_option("-3", "--min-mapq-bias", type=float, default=0,
                     help="Min P-value for mapQ bias (given PV4) "
                          "[%default]")
    group.add_option("-4", "--min-end-distance-bias", type=float, default=1e-4,
                     help="Min P-value for end distance bias (given PV4) "
                          "[%default]")
    parser.add_option_group(group)


def describe_filters(options):
    return {
        "HET": "Heterozygous SNPs observed on homozygous chromosome (e.g. chrX)",
        "q:%i" % options.min_quality: "Minimum Phred score recorded in the QUAL column",
        "f:%.4f" % options.min_allele_frequency:  "Minimum frequency of the alleles at heterozygous sites",
        "k": "SNPs without a most likely genotype (based on PL)",
        "Q:%i" % options.min_mapping_quality: "Minimum RMS mapping quality",
        "d:%i" % options.min_read_depth: "Minimum read depth",
        "D:%i" % options.max_read_depth: "Maximum read depth",
        "a:%i" % options.min_num_alt_bases: "Minimum number of alternative bases observed for variants",
        "w:%i" % options.min_distance_to_indels: "SNP within INT bp around a gap",
        "W:%i" % options.min_distance_between_indels: "Indel within INT bp of another indel",
        "1:%e" % options.min_strand_bias: "Min P-value for strand bias (given PV4)",
        "2:%e" % options.min_baseq_bias: "Min P-value for baseQ bias (given PV4)",
        "3:%e" % options.min_mapq_bias: "Min P-value for mapQ bias (given PV4)",
        "4:%e" % options.min_end_distance_bias: "Min P-value for end distance bias (given PV4)",
    }


def filter_vcfs(options, vcfs):
    vcfs  = iter(vcfs)
    chunk = collections.deque()
    filename = options.pileup
    min_freq = options.min_allele_frequency

    with AlleleFrequencies(filename, min_freq) as frequencies:
        while _read_chunk(vcfs, chunk):
            chunk = _filter_chunk(options, chunk, frequencies)
            for vcf in _trim_chunk(options, chunk):
                if vcf.filter == ".":
                    vcf.filter = "PASS"

                yield vcf


class AlleleFrequencies:
    VALID, INVALID, NA = range(3)

    def __init__(self, filename, min_freq):
        assert min_freq >= 0
        self._min_freq = min_freq
        self._handle   = None

        if filename and min_freq:
            self._handle = pysam.Tabixfile(filename)
            self.frequency_is_valid = self._frequency_is_valid
        else:
            self.frequency_is_valid = self._frequency_is_always_valid

    def _frequency_is_always_valid(self, contig, position, _ref, _first, _second):
        if self._min_freq:
            sys.stderr.write("WARNING: Multi-allelic SNP found at %s:%i, but --pileup has not been specified.\n" \
                             % (contig, position + 1))
        return self.VALID

    def _frequency_is_valid(self, contig, position, ref, first, second):
        assert self._handle

        if any((len(nt) > 1) for nt in (ref, first, second)):
            assert len(first) != len(second)
            first  = len(first) - len(ref)
            second = len(second) - len(ref)

        counts   = self._fetch(contig, position)
        n_first  = counts.get(first,  0)
        n_second = counts.get(second, 0)

        n_minor = min(n_first, n_second)
        n_major = max(n_first, n_second)
        if not n_major:
            return self.NA
        elif n_minor / float(n_minor + n_major) < self._min_freq:
            return self.INVALID
        return self.VALID

    def close(self):
        if self._handle:
            self._handle.close()
            self._handle = None
        self.frequency_is_valid = self._frequency_is_always_valid

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.close()

    def _fetch(self, contig, position):
        fetched = False
        for line in self._handle.fetch(contig, position, position + 1):
            fields, fetched = line.split("\t"), True
            assert len(fields) == 6
            break

        if not fetched:
            raise RuntimeError("Pileup did not contain position %s:%i, please rebuild." \
                               % (contig, position + 1))
        elif (fields[0] != contig) or (int(fields[1]) != position + 1):
            raise RuntimeError("Got wrong record (%s:%i vs %s:%s), is index corrupt?" \
                               % (contig, position + 1, fields[0], fields[1]))

        counts = {}
        bases  = list(fields[4][::-1].upper())
        ref    = fields[2]
        while bases:
            current = bases.pop()
            if current in "ACGTN":
                counts[current] = counts.get(current, 0) + 1
            elif current in ",.":
                counts[ref] = counts.get(ref, 0) + 1
            elif current in "+-":
                indel_length = [current]
                while bases[-1].isdigit():
                    indel_length.append(bases.pop())
                indel_length = int("".join(indel_length))

                for _ in xrange(abs(indel_length)):
                    bases.pop()

                counts[indel_length] = counts.get(indel_length, 0) + 1
            elif current == "*":
                counts[-1] = counts.get(-1, 0) + 1
            elif current == "^":
                bases.pop()
            elif current != "$":
                raise RuntimeError("Error parsing pileup (unexpected char '%s'): %s" \
                                       % (current, repr(line)))
        return counts


def _read_chunk(vcfs, chunk):
    try:
        while len(chunk) < _CHUNK_SIZE:
            chunk.append(vcfs.next())
    except StopIteration:
        chunk.append(None)

    return len(chunk) > 1


def _trim_chunk(options, chunk):
    min_distance = max(options.min_distance_between_indels,
                       options.min_distance_to_indels)

    if not chunk:
        return
    elif chunk[-1] is None:
        end_chr = "!@#$%^&*()_+"
        end_pos = _INF
        chunk.pop()
    else:
        end_chr = chunk[-1].contig
        end_pos = chunk[-1].pos

    while chunk:
        vcf = chunk[0]
        if  (vcf.contig == end_chr):
            # 'length' will become a too large value for heterozygous SNPs,
            # but it is faster than having to parse every position, and has
            # no effect on the final results.
            length = max(len(vcf.ref), len(vcf.alt))
            if (vcf.pos + length + min_distance) >= end_pos:
                break

        yield chunk.popleft()


def _group_indels_near_position(indels, distance):
    """Returns a dictionary of positions that are either directly covered by, or
    adjacent to indels, given some arbitrary distance. For each position, a list
    of adjacent/overlapping indels are provided."""
    positions = collections.defaultdict(list)
    if not distance:
        return positions

    for vcf in indels:
        # The number of bases covered (excluding the prefix)
        # For ambigious indels (e.g. in low complexity regions), this ensures
        # that the entire region is considered. Note that we do not need to
        # consider the alternative sequence(s)
        length = len(vcf.ref) - 1

        # Inclusive start/end positions for bases that should be blacklisted
        # Note that vcf.pos is the base just before the insertion/deletion
        start = vcf.pos + 1 - distance
        end   = vcf.pos + 1 + distance + length

        for position in xrange(start, end + 1):
            positions[position].append(vcf)

    return positions


def _select_best_indel(indels):
    """Select the highest quality indel, based on the quality,
    prefering low earlier positions above later positions in
    case of ties."""
    def _indel_by_quality_and_position(indel):
        # The negative position is used to select the first
        # of equally quality indels
        return (float(indel.qual), -indel.pos)

    return max(indels, key = _indel_by_quality_and_position)


def _filter_by_indels(options, chunk):
    """Filters a list of SNPs and Indels, such that no SNP is closer to
    an indel than the value set in options.min_distance_to_indels, and
    such that no two indels too close. If two or more indels are within
    this distance, the indel with the highest QUAL score is retained. When
    no unique highest QUAL score exists, an arbitrary indel is retained
    among those indels with the highest QUAL score. SNPs are filtered
    based on prefiltered Indels."""
    indels = [vcf for vcf in chunk if vcfwrap.is_indel(vcf)]

    distance_between = options.min_distance_between_indels
    indel_blacklist  = _group_indels_near_position(indels, distance_between)
    distance_to      = options.min_distance_to_indels
    snp_blacklist    = _group_indels_near_position(indels, distance_to)

    for vcf in chunk:
        if vcfwrap.is_indel(vcf):
            blacklisted = indel_blacklist.get(vcf.pos + 1, [vcf])
            if vcf is not _select_best_indel(blacklisted):
                _mark_as_filtered(vcf, "W:%i" % distance_between)
        elif (vcf.alt != ".") and (vcf.pos in snp_blacklist):
            # TODO: How to handle heterozygous SNPs near
            _mark_as_filtered(vcf, "w:%i" % distance_to)


def _filter_by_properties(options, vcfs, frequencies):
    """Filters a list of SNPs/indels based on the various properties recorded in
    the info column, and others. This mirrors most of the filtering carried out
    by vcfutils.pl varFilter."""
    for vcf in vcfs:
        if float(vcf.qual) < options.min_quality:
            _mark_as_filtered(vcf, "q:%i" % options.min_quality)

        properties = {}
        for field in vcf.info.split(";"):
            if "=" in field:
                key, value = field.split("=")
            else:
                key, value = field, None
            properties[key] = value

        read_depth = float(properties["DP"])
        if options.min_read_depth > read_depth:
            _mark_as_filtered(vcf, "d:%i" % options.min_read_depth)
        elif options.max_read_depth < read_depth:
            _mark_as_filtered(vcf, "D:%i" % options.max_read_depth)

        if "MQ" in properties:
            if float(properties["MQ"]) < options.min_mapping_quality:
                _mark_as_filtered(vcf, "Q:%i" % options.min_mapping_quality)

        if "PV4" in properties:
            pv4 = [float(value) for value in properties["PV4"].split(",")]
            if (pv4[0] < options.min_strand_bias):
                _mark_as_filtered(vcf, "1:%e" % options.min_strand_bias)
            if (pv4[1] < options.min_baseq_bias):
                _mark_as_filtered(vcf, "2:%e" % options.min_baseq_bias)
            if (pv4[2] < options.min_mapq_bias):
                _mark_as_filtered(vcf, "3:%e" % options.min_mapq_bias)
            if (pv4[3] < options.min_end_distance_bias):
                _mark_as_filtered(vcf, "4:%e" % options.min_end_distance_bias)

        if vcf.alt != ".":
            ref_fw, ref_rev, alt_fw, alt_rev = map(int, properties["DP4"].split(","))
            if (alt_fw + alt_rev) < options.min_num_alt_bases:
                _mark_as_filtered(vcf, "a:%i" % options.min_num_alt_bases)

            ml_genotype = vcfwrap.get_ml_genotype(vcf)
            if (ml_genotype == ("N", "N")) and not options.keep_ambigious_genotypes:
                # No most likely genotype
                _mark_as_filtered(vcf, "k")

            if (ml_genotype[0] != ml_genotype[1]):
                if vcf.contig in options.homozygous_chromosome:
                    _mark_as_filtered(vcf, "HET")

                # Filter by frequency of minor allele
                if vcf.ref in ml_genotype:
                    n_minor = min(ref_fw + ref_rev, alt_fw + alt_rev)
                    n_major = max(ref_fw + ref_rev, alt_fw + alt_rev)

                    if (n_minor / float(n_minor + n_major)) < options.min_allele_frequency:
                        _mark_as_filtered(vcf, "f:%.4f" % options.min_allele_frequency)
                else:
                    state = frequencies.frequency_is_valid(vcf.contig, vcf.pos, vcf.ref, *ml_genotype)
                    if state is frequencies.INVALID:
                        _mark_as_filtered(vcf, "f:%.4f" % options.min_allele_frequency)
                    elif state is frequencies.NA:
                        if _mark_as_filtered(vcf, "F:%.4f" % options.min_allele_frequency):
                            sys.stderr.write("WARNING: Could not determine allele-counts for SNP at %s:%s, filtering ...\n" % (vcf.contig, vcf.pos + 1))


def _filter_chunk(options, chunk, frequencies):
    at_end = False
    if chunk[-1] is None:
        at_end = True
        chunk.pop()

    _filter_by_indels(options, chunk)
    _filter_by_properties(options, chunk, frequencies)

    if at_end:
        chunk.append(None)
    return chunk


def _mark_as_filtered(vcf, filter_name):
    if vcf.filter in (".", "PASS"):
        vcf.filter = filter_name
        return True
    elif filter_name not in vcf.filter.split(";"):
        vcf.filter += ";" + filter_name
        return True



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Ignore "invalid" function names
# pylint: disable=C0103
# No need for docstrings
# pylint: disable=C0111
import copy

from nose.tools import \
    assert_equal, \
    assert_not_equal, \
    assert_raises

from paleomix.common.bedtools import \
    BEDRecord


###############################################################################
###############################################################################
# BEDRecord constructor

def test_bedrecord__constructor__defaults():
    record = BEDRecord()
    assert_equal(len(record), 0)
    assert_equal(str(record), "")
    assert_equal(repr(record), "BEDRecord()")


def test_bedrecord__constructor__empty_string():
    record = BEDRecord("")
    assert_equal(len(record), 0)
    assert_equal(str(record), "")
    assert_equal(repr(record), "BEDRecord()")


def test_bedrecord__constructor__3_fields():
    text = "my_contig\t12\t345"
    record = BEDRecord(text)

    assert_equal(len(record), 3)
    assert_equal(str(record), text)
    assert_equal(repr(record),
                 "BEDRecord(contig='my_contig', start=12, end=345)")


def test_bedrecord__constructor__6_fields():
    text = "my_contig\t12\t345\tmy_name\t-3\t-"
    record = BEDRecord(text)

    assert_equal(len(record), 6)
    assert_equal(str(record), text)
    assert_equal(repr(record),
                 "BEDRecord(contig='my_contig', start=12, "
                 "end=345, name='my_name', score=-3, strand='-')")


def test_bedrecord__constructor__extra_fields():
    text = "my_contig\t12\t345\tmy_name\t-3\t-\tfoo\tbar"
    record = BEDRecord(text)

    assert_equal(len(record), 8)
    assert_equal(str(record), text)
    assert_equal(repr(record),
                 "BEDRecord(contig='my_contig', start=12, "
                 "end=345, name='my_name', score=-3, strand='-', "
                 "'foo', 'bar')")


###############################################################################
###############################################################################
# BEDRecord accessors

def test_bedrecord__accessors__3_fields():
    record = BEDRecord("my_contig\t12\t345")

    assert_equal(record.contig, "my_contig")
    assert_equal(record.start, 12)
    assert_equal(record.end, 345)
    assert_raises(IndexError, lambda: record.name)
    assert_raises(IndexError, lambda: record.score)
    assert_raises(IndexError, lambda: record.strand)


def test_bedrecord__accessors__6_fields():
    record = BEDRecord("my_contig\t12\t345\tmy_name\t-3\t-")

    assert_equal(record.contig, "my_contig")
    assert_equal(record.start, 12)
    assert_equal(record.end, 345)
    assert_equal(record.name, "my_name")
    assert_equal(record.score, -3)
    assert_equal(record.strand, "-")


def test_bedrecord__accessors__extra_fields():
    text = "my_contig\t12\t345\tmy_name\t-3\t-\tfoo\tbar"
    record = BEDRecord(text)

    assert_equal(record[6], "foo")
    assert_equal(record[7], "bar")


def test_bedrecord__accessors__6_fields__getitem():
    record = BEDRecord("my_contig\t12\t345\tmy_name\t-3\t-")

    assert_equal(record[0], "my_contig")
    assert_equal(record[1], 12)
    assert_equal(record[2], 345)
    assert_equal(record[3], "my_name")
    assert_equal(record[4], -3)
    assert_equal(record[5], "-")


def test_bedrecord__setters__3_fields():
    record = BEDRecord("my_contig\t12\t345")

    record.contig = "chrZ"
    assert_equal(record.contig, "chrZ")

    record.end += 20
    assert_equal(record.end, 365)

    assert_equal(str(record), "chrZ\t12\t365")
    assert_equal(repr(record),
                 "BEDRecord(contig='chrZ', start=12, end=365)")


def test_bedrecord__setters__type_errors():
    record = BEDRecord("my_contig\t12\t345\tname\t0\t+")

    assert_raises(ValueError, lambda: setattr(record, "contig", 17))
    assert_raises(ValueError, lambda: setattr(record, "start", "foo"))
    assert_raises(ValueError, lambda: setattr(record, "end", "foo"))
    assert_raises(ValueError, lambda: setattr(record, "name", 17.3))
    assert_raises(ValueError, lambda: setattr(record, "score", "foo"))
    assert_raises(ValueError, lambda: setattr(record, "strand", "foo"))


def test_bedrecord__setters__unset_fields__at_end():
    record = BEDRecord("my_contig\t12\t345")

    record.name = "my_region"
    assert_equal(record.name, "my_region")

    record.score = -13
    assert_equal(record.score, -13)

    record.strand = '-'
    assert_equal(record.strand, '-')

    assert_equal(str(record), "my_contig\t12\t345\tmy_region\t-13\t-")
    assert_equal(repr(record),
                 "BEDRecord(contig='my_contig', start=12, end=345, "
                 "name='my_region', score=-13, strand='-')")


def test_bedrecord__setters__unset_fields__after_end():
    record = BEDRecord("")
    record.strand = "-"
    assert_equal(str(record), "\t0\t0\t\t0\t-")

    record = BEDRecord("my_name")
    record.strand = "-"
    assert_equal(str(record), "my_name\t0\t0\t\t0\t-")

    record = BEDRecord("my_name\t17")
    record.strand = "-"
    assert_equal(str(record), "my_name\t17\t0\t\t0\t-")

    record = BEDRecord("my_name\t17\t258")
    record.strand = "-"
    assert_equal(str(record), "my_name\t17\t258\t\t0\t-")

    record = BEDRecord("my_name\t17\t258\tregion")
    record.strand = "-"
    assert_equal(str(record), "my_name\t17\t258\tregion\t0\t-")

    record = BEDRecord("my_name\t17\t258\tregion\t33")
    record.strand = "-"
    assert_equal(str(record), "my_name\t17\t258\tregion\t33\t-")

    record = BEDRecord("my_name\t17\t258\tregion\t33\t+")
    record.strand = "-"
    assert_equal(str(record), "my_name\t17\t258\tregion\t33\t-")


def test_bedrecord__cmp():
    record_1_txt = "my_contig\t12\t345\tmy_name\t-3\t-\tfoo"
    record_1 = BEDRecord(record_1_txt)
    record_2 = BEDRecord("chrZ\t132\t4345\tchrZ_region\t0\t+\tbar")

    for idx in xrange(len(record_2)):
        record_tmp = BEDRecord(record_1_txt)
        assert_equal(record_1, record_tmp)
        record_tmp[idx] = record_2[idx]
        assert_not_equal(record_1, record_tmp)
        record_tmp[idx] = record_1[idx]
        assert_equal(record_1, record_tmp)


###############################################################################
###############################################################################

def test_bedrecord__copy():
    record_1_txt = "my_contig\t12\t345\tmy_name\t-3\t-"
    record_1 = BEDRecord(record_1_txt)
    record_2 = copy.copy(record_1)
    record_2.name = "my_clone"

    assert_equal(str(record_1), record_1_txt)
    assert_equal(str(record_2), "my_contig\t12\t345\tmy_clone\t-3\t-")


def test_bedrecord__deepcopy():
    record_1_txt = "my_contig\t12\t345\tmy_name\t-3\t-"
    record_1 = BEDRecord(record_1_txt)
    record_1[6] = ["foo"]
    record_2 = copy.deepcopy(record_1)
    record_2[6][0] = "bar"

    assert_equal(str(record_1), record_1_txt + "\t['foo']")
    assert_equal(str(record_2), record_1_txt + "\t['bar']")



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Various functions relating to DNA sequence manipulation."""

import itertools


# Pairs of complementary bases and ambigious basees
_COMPL = [ "AT", "CG",
           "NN", "RY",
           "KM", "SS",
           "WW", "BV",
           "DH", "XX" ]
_COMPL_TABLE = ["N"] * 256
for (_a, _b) in _COMPL :
    # Complement both upper/lower-case bases
    for _func in (str.upper, str.lower):
        _COMPL_TABLE[ord(_func(_a))] = _func(_b)
        _COMPL_TABLE[ord(_func(_b))] = _func(_a)
_COMPL_TABLE = "".join(_COMPL_TABLE)



# Table of nt codes (IUPAC codes) used to encode (ambigious) bases:
#   Nomenclature for incompletely specified bases in nucleic acid sequences.
#   Recommendations 1984. J Biol Chem. 1986 Jan 5;261(1):13-7.
#   PubMed PMID: 2416744.
NT_CODES = [
    ["A", "A"],
    ["C", "C"],
    ["G", "G"],
    ["T", "T"],
    ["N", "N"],
    ["R", "AG"],
    ["Y", "CT"],
    ["K", "GT"],
    ["M", "AC"],
    ["S", "CG"],
    ["W", "AT"],
    ["B", "CGT"],
    ["D", "AGT"],
    ["H", "ACT"],
    ["V", "ACG"],
    ["N", "ACGT"]]

_NT_CODES_TABLE = {}
for (_abr, _nts) in NT_CODES:
    _NT_CODES_TABLE[frozenset(_nts)] = _abr
    _NT_CODES_TABLE[frozenset(_nts + ",")] = _abr

NT_CODES = dict(NT_CODES)


CODONS = {
    "+": {
        "TTT": "Phe",  "TCT": "Ser",  "TAT": "Tyr",  "TGT": "Cys",
        "TTC": "Phe",  "TCC": "Ser",  "TAC": "Tyr",  "TGC": "Cys",
        "TTA": "Leu",  "TCA": "Ser",  "TAA": "Stop", "TGA": "Stop",
        "TTG": "Leu",  "TCG": "Ser",  "TAG": "Stop", "TGG": "Trp",

        "CTT": "Leu",  "CCT": "Pro",  "CAT": "His",  "CGT": "Arg",
        "CTC": "Leu",  "CCC": "Pro",  "CAC": "His",  "CGC": "Arg",
        "CTA": "Leu",  "CCA": "Pro",  "CAA": "Gln",  "CGA": "Arg",
        "CTG": "Leu",  "CCG": "Pro",  "CAG": "Gln",  "CGG": "Arg",

        "ATT": "Ile",  "ACT": "Thr",  "AAT": "Asn",  "AGT": "Ser",
        "ATC": "Ile",  "ACC": "Thr",  "AAC": "Asn",  "AGC": "Ser",
        "ATA": "Ile",  "ACA": "Thr",  "AAA": "Lys",  "AGA": "Arg",
        "ATG": "Met",  "ACG": "Thr",  "AAG": "Lys",  "AGG": "Arg",

        "GTT": "Val",  "GCT": "Ala",  "GAT": "Asp",  "GGT": "Gly",
        "GTC": "Val",  "GCC": "Ala",  "GAC": "Asp",  "GGC": "Gly",
        "GTA": "Val",  "GCA": "Ala",  "GAA": "Glu",  "GGA": "Gly",
        "GTG": "Val",  "GCG": "Ala",  "GAG": "Glu",  "GGG": "Gly"},
    "-": {}}


def complement(sequence):
    """Returns the complement of a DNA sequence (string)."""
    return sequence.translate(_COMPL_TABLE)


def reverse_complement(sequence):
    """Returns the reverse complement of a DNA sequence."""
    return complement(sequence)[::-1]


for _codon, _aa in CODONS["+"].iteritems():
    assert not set(_codon) - set("ACGT")
    CODONS["-"][reverse_complement(_codon)] = _aa

assert len(CODONS["+"]) == 64
assert len(CODONS["-"]) == 64


def encode_genotype(nucleotides):
    """Parses a string representing a set of nucleotides observed at a loci,
    and returns the corresponding IUPAC code. Commas are allowed, but are
    simply ignored if found in the string. Does not handle lower-case
    nucleotides, due to lack of clear criteria for mixed case input.
    See e.g. http://www.ebi.ac.uk/2can/tutorials/aa.html"""
    try:
        return _NT_CODES_TABLE[frozenset(nucleotides)]
    except KeyError:
        raise ValueError("Invalid input for 'encode_genotype': %s" % (repr(nucleotides), ))



def count_nts(sequence):
    """Given a nucleotide sequence (str), this function returns
    the number of each type of nucleotide representable using
    IUPAC codes. The sequence must not contain non-IUPAC
    nucleotides, or other annotation. IUPAC nucleotides are
    handled in a case-insensitive manner."""
    counts = {}
    sequence = sequence.upper()
    for nucleotide in NT_CODES:
        count = sequence.count(nucleotide)
        if count:
            counts[nucleotide] = count

    if len(sequence) != sum(counts.itervalues()):
        raise ValueError("Sequence contains non-(IUPAC-)nucleotides: %s" % \
                         ", ".join(set(sequence) - set(counts)))

    return counts


def count_gc_diploid(sequence):
    """Given a sequence, this function returns a tuple containing the
    the total number of bases that were G/C, as well as the total number
    of bases. The sequence is assumed to represent a diploid genome, with
    the total number of bases being twice the sequence length, and
    hence IUPAC codes representing one of or both of G/C are treated as
    reflecting both strands. Thus R counts for 1, while S counts for 2.

    The sequence must only contain valid IUPAC codes, and no other
    form of annotation. Both uppercase/lowercase G/Cs are counted.
    Ambigious site (n/N) are not counted, neither in the number of G/C,
    nor in the total number of bases."""
    total_nts = total_gc = 0
    counts = count_nts(sequence)
    for (code, count) in counts.iteritems():
        value = 0
        if code == "N":
            continue
        elif code in "CGcg":
            value = 2
        else:
            code_represents = NT_CODES[code]
            if (len(code_represents) > 2) and (code != 'N'):
                raise ValueError("calculate_gcp assumes diploid genome, nt code for tri-valued SNP observed: " + code)

            if 'G' in code_represents:
                value += 1
            if 'C' in code_represents:
                value += 1

        total_nts += count * 2
        total_gc += count * value

    return (total_gc, total_nts)


def split(sequence, split_by = "123"):
    """Splits a sequence by position, as specified by the 'split_by' parameter. By
    default, the function will split by codon position, and return a dictionary
    containing the keys '1', '2' and '3'.

    The 'split_by' parameter may contain any non-zero number of values, which must
    however be hashable. If a value is specified multiple times, then those positions
    are interleaved (e.g. split_by = "112" returns the first two positions in a codon
    as one sequence, as well as the last positions as one sequence."""
    if not split_by:
        raise TypeError("No partitions to split by specified")

    results = dict((key, []) for key in split_by)
    keys = itertools.chain(itertools.cycle(split_by))
    for (key, nucleotide) in itertools.izip(keys, sequence):
        results[key].append(nucleotide)

    for key in results:
        results[key] = "".join(results[key])

    return results



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import print_function

import sys
import errno
import optparse
import fileinput

import pysam

import paleomix.common.vcffilter as vcffilter


def _read_files(filenames, args):
    in_header = True
    has_filters = False
    vcf_parser = pysam.asVCF()
    for line in fileinput.input(filenames):
        if not line.startswith("#"):
            in_header = False
            line = line.rstrip("\n\r")
            yield vcf_parser(line, len(line))
        elif in_header:
            if not (line.startswith("##") or has_filters):
                has_filters = True
                for item in sorted(vcffilter.describe_filters(args).items()):
                    print('##FILTER=<ID=%s,Description="%s">' % item)

            print(line, end="")


def main(argv):
    parser = optparse.OptionParser("paleomix vcf_filter [options] [in1.vcf, ...]")
    vcffilter.add_varfilter_options(parser)
    (opts, args) = parser.parse_args(argv)

    if (not args or "-" in args) and sys.stdin.isatty():
        parser.error("STDIN is a terminal, terminating!")

    try:
        for vcf in vcffilter.filter_vcfs(opts, _read_files(args, opts)):
            print(vcf)
    except IOError, error:
        # Check for broken pipe (head, less, etc).
        if error.errno != errno.EPIPE:
            raise


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import argparse
import collections
import itertools
import os
import random
import sys
import tarfile

import pysam

from paleomix.common.sequences import NT_CODES
from paleomix.common.sampling import reservoir_sampling

import paleomix.common.bamfiles as bamtools
import paleomix.common.fileutils as fileutils
import paleomix.tools.zonkey.common as common
import paleomix.tools.zonkey.database as database


_TRANSITIONS = frozenset((('C', 'T'), ('T', 'C'),
                          ('G', 'A'), ('A', 'G')))


def _filter_records(handle, flags=bamtools.EXCLUDED_FLAGS):
    for record in handle:
        if not record.flag & flags:
            yield record


class DownsampledBAM(object):
    def __init__(self, handle, downsample, included_references):
        self._records = collections.defaultdict(list)

        references = handle.references
        if len(references) != len(included_references):
            raise ValueError("Length of 'included_references' must match the "
                             "number of references in BAM file.")

        records = _filter_records(handle)
        for record in reservoir_sampling(records, downsample):
            key = references[record.tid]
            self._records[key].append(record)

        self.references = references

        self._records = dict(self._records)
        for value in self._records.values():
            value.sort(key=lambda rec: rec.pos)

    def fetch(self, chrom):
        return self._records.get(chrom, ())


class GenotypeSites(object):
    def __init__(self, records):
        last_chrom = None
        sites = []

        # chrom, pos, ref, ..., nucleotides
        for chrom, pos, line in records:
            # Convert pos from 1-based to 0-based (same as BAM.pos)
            sites.append((int(pos) - 1, line, []))

            assert last_chrom is None or chrom == last_chrom, \
                (chrom, last_chrom)
            last_chrom = chrom

        sites.sort()
        self._sites = collections.deque(sites)

    def process(self, records, statistics):
        count_used = 0
        count_total = 0
        sites = self._sites
        for record in records:
            count_total += 1

            # TODO: Check sorted
            while sites and sites[0][0] < record.pos:
                yield sites.popleft()

            sequence = record.seq
            sites_iter = iter(sites).next
            alignment_iter = iter(record.get_aligned_pairs()).next
            alignment_end = record.aend

            try:
                read_used = False
                site_pos, _, nucleotides = sites_iter()
                query_pos, ref_pos = alignment_iter()

                while alignment_end > site_pos:
                    if ref_pos is None or query_pos is None or site_pos > ref_pos:
                        query_pos, ref_pos = alignment_iter()
                    elif site_pos < ref_pos:
                        site_pos, _, nucleotides = sites_iter()
                    else:
                        assert ref_pos == site_pos, (ref_pos, site_pos)
                        nucleotide = sequence[query_pos]
                        if nucleotide != "N":
                            nucleotides.append(nucleotide)

                        read_used = True
                        query_pos, ref_pos = alignment_iter()
                        site_pos, _, nucleotides = sites_iter()
            except StopIteration:
                if not sites:
                    break
            finally:
                if read_used:
                    count_used += 1

        while sites:
            yield sites.popleft()

        statistics["n_reads"] += count_total
        statistics["n_reads_used"] += count_used

    @classmethod
    def _parse_cigar(cls, record):
        seq_iter = iter(record.seq)
        seq_pos = record.pos
        for key, value in record.cigar:
            if key in (0, 7, 8):  # M, =, X
                for _ in xrange(value):
                    yield seq_pos, seq_iter.next()
                    seq_pos += 1
            elif key in (1, 4):  # I
                for _ in xrange(value):
                    seq_iter.next()
            elif key in (2, 3, ):  # D, N
                seq_pos += value
            elif key in (5, 6):  # H, P
                pass
            else:
                raise ValueError(record.cigar)


class GenotypeReader(object):
    def __init__(self, filename):
        self._tar_handle = tarfile.open(filename)
        self._handle = self._tar_handle.extractfile("genotypes.txt")
        self._header = self._handle.readline().rstrip('\r\n').split('\t')
        self.samples = self._header[-1].split(';')

    def __iter__(self):
        for chrom, records in itertools.groupby(self._read_records(),
                                                lambda rec: rec[0]):
            sys.stderr.write("Reading contig %r information ...\n" % (chrom,))
            yield chrom, GenotypeSites(records)

    def _read_records(self):
        for line in self._handle:
            yield line.rstrip().split('\t', 2)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._handle.close()
        self._tar_handle.close()


def process_record(chrom, pos, line, nucleotides, statistics,
                   out_incl_ts=sys.stdout,
                   out_excl_ts=sys.stdout):
    if not nucleotides:
        return
    elif len(nucleotides) == 1:
        # Avoid unnessary random() call in 'random.choice'
        nucleotide = nucleotides[0]
    else:
        nucleotide = random.choice(nucleotides)

    # Fields are expected to contain at least 2 columns, the first being the
    # reference nucleotide, and the last being the sample genotypes
    _, encoded_genotypes = line.strip().rsplit('\t', 1)

    genotypes = []
    for encoded_nucleotide in encoded_genotypes:
        decoded_nucleotides = NT_CODES.get(encoded_nucleotide, ())
        if len(decoded_nucleotides) == 1:
            genotypes.append(decoded_nucleotides)
            genotypes.append(decoded_nucleotides)
        elif len(decoded_nucleotides) == 2:
            genotypes.extend(decoded_nucleotides)
        else:
            raise ValueError('Invalid nucleotide, not bi-allelic: %r'
                             % (encoded_nucleotide,))

    if nucleotide not in genotypes:
        # Exclude SNPs not observed in the reference panel
        return

    # Convert from 0-based to 1-based
    pos += 1

    # Chromosome, SNP identifier, (dummy) pos in (centi)Morgans, position
    output = [chrom, "chr{}_{}".format(chrom, pos), "0", str(pos)]
    output.extend(genotypes)
    output.append(nucleotide)
    output.append(nucleotide)
    output = " ".join(output)

    statistics["n_sites_incl_ts"] += 1
    out_incl_ts.write("{}\n".format(output))

    if tuple(set(genotypes)) not in _TRANSITIONS:
        statistics["n_sites_excl_ts"] += 1
        out_excl_ts.write("{}\n".format(output))


def write_tfam(filename, data, samples, bam_sample):
    with open(filename, "w") as handle:
        for key in samples:
            row = data.samples[key]
            sex = {"MALE": 1, "FEMALE": 2, "NA": 0}[row["Sex"].upper()]
            # Family, Individual, Paternal ID, Maternal ID, Sex, Phenotype
            handle.write("{0} {0} 0 0 {1} -9\n".format(key, sex))

        handle.write("{0} {0} 0 0 0 -9\n".format(bam_sample))


def write_summary(args, filename, statistics):
    with open(filename, "w") as handle:
        handle.write("name: %s\n" % (args.name,))
        handle.write("filename: %s\n" % (os.path.abspath(args.bam),))

        for key in ("n_reads", "n_reads_used",
                    "n_sites_incl_ts", "n_sites_excl_ts"):
            handle.write("%s: %s\n" % (key, statistics.get(key, "MISSING")))


def process_bam(args, data, bam_handle):
    raw_references = bam_handle.references
    references = map(common.contig_name_to_plink_name, raw_references)

    if args.downsample:
        sys.stderr.write("Downsampling to at most %i BAM records ...\n"
                         % (args.downsample))
        bam_handle = DownsampledBAM(bam_handle, args.downsample, references)

    statistics = {"n_reads": 0,
                  "n_reads_used": 0,
                  "n_sites_incl_ts": 0,
                  "n_sites_excl_ts": 0}

    fileutils.make_dirs(args.root)

    with open(os.path.join(args.root, 'incl_ts.tped'), 'w') as output_incl:
        with open(os.path.join(args.root, 'excl_ts.tped'), 'w') as output_excl:
            with GenotypeReader(args.data) as reader:
                for ref, sites in reader:
                    raw_ref = raw_references[references.index(ref)]

                    sys.stderr.write("Reading %r from BAM ...\n" % (raw_ref,))
                    raw_sites = bam_handle.fetch(raw_ref)
                    for pos, line, nucleotides in sites.process(raw_sites,
                                                                statistics):
                        process_record(ref, pos, line, nucleotides,
                                       out_incl_ts=output_incl,
                                       out_excl_ts=output_excl,
                                       statistics=statistics)

                write_summary(args, os.path.join(args.root, "common.summary"),
                              statistics=statistics)
                write_tfam(os.path.join(args.root, "common.tfam"),
                           data, reader.samples, args.name)


def parse_args(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument('root')
    parser.add_argument('data')
    parser.add_argument('bam')
    parser.add_argument('--seed', default=7913, type=int)  # FIXME
    parser.add_argument('--downsample', type=int)
    parser.add_argument('--name', default="Sample")

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)
    random.seed(args.seed)

    print "Reading reference information from %r ..." % (args.data,)

    try:
        data = database.ZonkeyDB(args.data)
    except database.ZonkeyDBError, error:
        sys.stderr.write("Error reading database file %r:\n%s\n"
                         % (args.data, error))
        return 1

    with pysam.Samfile(args.bam) as bam_handle:
        bam_info = data.validate_bam_handle(bam_handle)
        if not (bam_info or bam_info.is_nuclear):
            # TODO: Verbose error
            return 1

        process_bam(args, data, bam_handle)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# pylint: disable=W0212

from __future__ import print_function

import os
import sys
import types
import subprocess


def _is_cls(obj, *cls_names):
    return obj.__class__.__name__ in cls_names

def _get_pipe_name(files, pipe):
    if pipe in files:
        return pipe.split("_")[-1] + " "
    return pipe.split("_")[-1] + "*"

def _get_pipe_file(files, pipe):
    pipe_filename = files.get(pipe)
    if pipe_filename:
        return pipe_filename
    return files.get("TEMP_%s" % (pipe,))

def _describe_cls(atomiccmd):
    if _is_cls(atomiccmd, "ParallelCmds"):
        return "Parallel commands"
    elif _is_cls(atomiccmd, "SequentialCmds"):
        return "Sequential commands"
    assert False # pragma: no coverage


def _collect_stats(atomiccmd, stats):
    assert atomiccmd not in stats["id"]

    if _is_cls(atomiccmd, "AtomicCmd"):
        stats["id"][atomiccmd] = len(stats["id"])
        pipe   = _get_pipe_file(atomiccmd._files, "IN_STDIN")
        if _is_cls(pipe, "AtomicCmd"):
            stats["pipe"][pipe] = atomiccmd
    elif _is_cls(atomiccmd, "ParallelCmds", "SequentialCmds"):
        for subcmd in atomiccmd._commands:
            _collect_stats(subcmd, stats)
    else:
        assert False # pragma: no coverage

    return stats


def _build_status(atomiccmd, _stats, indent, lines):
    prefix = " " * indent + "Status  = "
    if atomiccmd._proc:
        if atomiccmd.ready():
            return_code = tuple(atomiccmd.join())
            if isinstance(return_code[0], types.StringTypes):
                lines.append(prefix + "Terminated with signal %s" % return_code)
            else:
                lines.append(prefix + "Exited with return-code %i" % return_code)
        else:
            lines.append(prefix + "Running ...")


def _build_stdin(atomiccmd, files, stats, indent, lines):
    pipe_name = _get_pipe_name(files, "IN_STDIN")
    pipe   = _get_pipe_file(files, "IN_STDIN")
    prefix = "%s%s  = " % (" " * indent, pipe_name)
    if pipe and pipe in stats["id"]:
        lines.append("%s<%02i>" % (prefix, stats["id"][pipe],))
    elif isinstance(pipe, types.StringTypes):
        if atomiccmd._set_cwd and (pipe_name == "STDIN*"):
            pipe = os.path.basename(pipe)
        lines.append("%s'%s'" % (prefix, pipe))
    elif pipe:
        lines.append("%s<PIPE>" % (prefix,))


def _build_out_pipe(atomiccmd, files, stats, indent, lines, pipe):
    pipe_name = _get_pipe_name(files, pipe)
    prefix = "%s%s = " % (" " * indent, pipe_name)

    if (atomiccmd in stats["pipe"]) and (pipe == "OUT_STDOUT"):
        pipe = stats["pipe"].get(atomiccmd)
        lines.append("%s<%02i>" % (prefix, stats["id"][pipe],))
        return

    filename = _get_pipe_file(files, pipe)
    if filename is not subprocess.PIPE:
        lines.append("%s'%s'" % (prefix, filename))
    else:
        lines.append("%s<PIPE>" % (prefix,))


def _build_cwd(atomiccmd, indent, lines):
    prefix = " " * indent + "CWD     = "
    if atomiccmd._temp:
        if atomiccmd._set_cwd:
            lines.append("%s'%s'" % (prefix, atomiccmd._temp,))
        else:
            lines.append("%s'%s'" % (prefix, os.getcwd()))
    elif atomiccmd._set_cwd:
        lines.append("%s'%s'" % (prefix, "${TEMP_DIR}"))


def _pformat(atomiccmd, stats, indent, lines, include_prefix = True):
    s_prefix     = ""
    if include_prefix:
        s_prefix = " " * indent + "- "
        if _is_cls(atomiccmd, "AtomicCmd"):
            cmd_id = stats["id"][atomiccmd]
            s_prefix += "<%02i> " % (cmd_id,)
    s_prefix_len = len(s_prefix)

    if _is_cls(atomiccmd, "AtomicCmd"):
        temp = "" if atomiccmd._set_cwd else (atomiccmd._temp or "${TEMP_DIR}")
        files = atomiccmd._generate_filenames(atomiccmd._files, temp)

        c_prefix = s_prefix + "Command = "
        for line in _pformat_list(atomiccmd._generate_call(temp)).split("\n"):
            lines.append("%s%s" % (c_prefix, line))
            c_prefix = " " * len(c_prefix)

        if not s_prefix_len:
            s_prefix_len += 1

        _build_status(atomiccmd,   stats, s_prefix_len, lines)
        _build_stdin(atomiccmd,    files, stats, s_prefix_len, lines)
        _build_out_pipe(atomiccmd, files, stats, s_prefix_len, lines, "OUT_STDOUT")
        _build_out_pipe(atomiccmd, files, stats, s_prefix_len, lines, "OUT_STDERR")
        _build_cwd(atomiccmd,                    s_prefix_len, lines)
    elif _is_cls(atomiccmd, "ParallelCmds", "SequentialCmds"):
        lines.append("%s%s:" % (s_prefix, _describe_cls(atomiccmd)))
        for subcmd in atomiccmd._commands:
            _pformat(subcmd, stats, s_prefix_len + 2, lines)
    else:
        assert False # pragma: no coverage


def _pformat_list(lst, width = 80):
    """Return a printable representation of a list, where line-breaks
    are inserted between items to minimize the number of lines with a
    width greater than 'width'. Very long items may cause this maximum
    to be exceeded."""
    result = [[]]
    current_width = 0
    for item in map(repr, lst):
        if current_width + len(item) + 2 > width:
            if not result[-1]:
                result[-1] = [item]
                current_width = len(item) + 2
            else:
                result.append([item])
                current_width = len(item) + 2
        else:
            result[-1].append(item)
            current_width += len(item) + 2

    return "[%s]" % (",\n ".join(", ".join(line) for line in result))


def pformat(atomiccmd):
    """Returns a human readable description of an Atomic Cmd or Atomic Set
    of commands. This is currently equivalent to str(cmd_obj)."""
    if not _is_cls(atomiccmd, "AtomicCmd", "ParallelCmds", "SequentialCmds"):
        raise TypeError("Invalid type in pformat: %r" % atomiccmd.__class__.__name__)

    lines = []
    stats = _collect_stats(atomiccmd, {"id" : {}, "pipe" : {}})
    _pformat(atomiccmd, stats, 0, lines, False)
    return "<%s>" % "\n".join(lines)


def pprint(atomiccmd, out = sys.stdout):
    """Prints a human readable description of an Atomic Cmd or Atomic Set
    of commands. This is currently equivalent to print(str(cmd_obj), ...)."""
    print(pformat(atomiccmd), file = out)




#!/usr/bin/env python

"""
Stripped down version of 'FilterUniqueBAM' by
:Author: Martin Kircher
:Contact: Martin.Kircher@eva.mpg.de
:Date: *08.10.2011
:Type: tool
:Input: BAM
:Output: BAM

Mark/Filter PCR duplicates for merged PE reads Reads BAM
from STDIN and writes BAM to STDOUT. All non-collapsed reads
as well as secondary/chinermic alignments, reads that have
failed QC and unmmaped reads, are written to STDOUT as is.

The input is assumed to be sorted by coordinates, and this
order is preservered, though individual reads at the same
position may be re-arranged).
"""

import sys
import pysam
from argparse import ArgumentParser


def calc_consensus(reads):
    count = len(reads)
    outread = None
    maxsumqual = 0
    for read in reads:
        nsum = sum(map(ord, read.qual))
        if nsum > maxsumqual:
            outread = read
            maxsumqual = nsum

        # LOOK FOR PREVIOUS PCR DUPLICATE COUNTS
        for key, value in read.tags:
            if key == "XP":
                count += value

    if not outread.tags:
        outread.tags = [("XP", count)]
    else:
        outread.tags = outread.tags + [("XP", count)]

    return outread


def get_consensus_se(reads):
    # DETERMINE MOST FREQUENT CIGAR LINE
    by_cigar = {}
    cigar_count = {}
    for read in reads:
        tcigar = tuple(read.cigar)
        if tcigar in by_cigar:
            cigar_count[tcigar] += 1
            by_cigar[tcigar].append(read)
        else:
            cigar_count[tcigar] = 1
            by_cigar[tcigar] = [read]

    to_sort = [(y, -len(str(x)), x) for (x, y) in cigar_count.iteritems()]
    to_sort.sort()
    selcigar = to_sort[-1][-1]
    reads = by_cigar[selcigar]

    return calc_consensus(reads)


def write_consensus_se(outfile, reads, remove_duplicates):
    consensus = get_consensus_se(reads)
    for read in reads:
        read.is_duplicate = (read is not consensus)
        if not (read.is_duplicate and remove_duplicates):
            outfile.write(read)


def _flush_buffer(outfile, curvariants, remove_duplicates):
    for value in curvariants.itervalues():
        write_consensus_se(outfile, value[0], remove_duplicates)
    curvariants.clear()


_FILTERED_FLAGS = 0x1     # PE reads
_FILTERED_FLAGS |= 0x4    # Unmapped
_FILTERED_FLAGS |= 0x100  # Secondary alignment
_FILTERED_FLAGS |= 0x200  # Failed QC
_FILTERED_FLAGS |= 0x800  # Chimeric alignment


def parse_args(argv):
    usage = "paleomix rmdup_collapsed [options] < sorted.bam > out.bam"
    parser = ArgumentParser(usage=usage)
    parser.add_argument("input", default="-", help="BAM file.", nargs="?")
    parser.add_argument("--remove-duplicates",
                        help="Remove duplicates from output; by default "
                             "duplicates are only flagged (flag = 0x400).",
                        default=False, action="store_true")

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)

    if args.input == "-" and sys.stdin.isatty():
        sys.stderr.write("STDIN is a terminal, terminating!\n")
        return 1
    elif sys.stdout.isatty():
        sys.stderr.write("STDOUT is a terminal, terminating!\n")
        return 1

    with pysam.Samfile(args.input, "rb") as infile:
        with pysam.Samfile("-", "wb", template=infile) as outfile:
            curpos = None
            curvariants = {}
            for (read_num, read) in enumerate(infile):
                if curpos and ((read.tid, read.pos) != curpos):
                    # Sort order is defined as ascending 'tid's and positions
                    if curpos > (read.tid, read.pos) and not read.is_unmapped:
                        sys.stderr.write("ERROR: Input file does not appear "
                                         "to be sorted by coordinates at "
                                         "record %i, aborting ...\n"
                                         % (read_num,))
                        return 1

                    _flush_buffer(outfile, curvariants,
                                  args.remove_duplicates)
                    curpos = None

                is_filtered = read.flag & _FILTERED_FLAGS
                is_collapsed = read.qname.startswith("M_")
                if is_filtered or not (read.qual and is_collapsed):
                    outfile.write(read)
                    continue

                curpos = (read.tid, read.pos)
                nkey = (read.is_reverse, read.pos, read.alen)
                if nkey in curvariants:
                    curvariants[nkey][0].append(read)
                    curvariants[nkey][1] += 1
                else:
                    curvariants[nkey] = [[read], 1]

            _flush_buffer(outfile, curvariants, args.remove_duplicates)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
